"""
This module queries process memory allocation metrics from the operating
system. It provides a platform independent layer to get the amount of virtual
and physical memory allocated to the Python process.

Different mechanisms are implemented: Either the process stat file is read
(Linux), the `ps` command is executed (BSD/OSX/Solaris) or the resource module
is queried (Unix fallback). On Windows try to use the win32 module if
available. If all fails, return 0 for each attribute.

Windows without the win32 module is not supported.

    >>> from pympler.process import ProcessMemoryInfo
    >>> pmi = ProcessMemoryInfo()
    >>> print ("Virtual size [Byte]: " + str(pmi.vsz)) # doctest: +ELLIPSIS
    Virtual size [Byte]: ...
"""

import logging
import threading

try:
    from thread import get_ident
except ImportError:
    from _thread import get_ident  # NOQA

from os import getpid
from subprocess import Popen, PIPE

from pympler.util.stringutils import pp

try:
    from resource import getpagesize as _getpagesize
except ImportError:
    _getpagesize = lambda: 4096


class _ProcessMemoryInfo(object):
    """Stores information about various process-level memory metrics. The
    virtual size is stored in attribute `vsz`, the physical memory allocated to
    the process in `rss`, and the number of (major) pagefaults in `pagefaults`.
    On Linux, `data_segment`, `code_segment`, `shared_segment` and
    `stack_segment` contain the number of Bytes allocated for the respective
    segments.  This is an abstract base class which needs to be overridden by
    operating system specific implementations. This is done when importing the
    module.
    """

    pagesize = _getpagesize()

    def __init__(self):
        self.pid = getpid()
        self.rss = 0
        self.vsz = 0
        self.pagefaults = 0
        self.os_specific = []

        self.data_segment = 0
        self.code_segment = 0
        self.shared_segment = 0
        self.stack_segment = 0

        self.available = self.update()

    def __repr__(self):
        return "<%s vsz=%d rss=%d>" % (self.__class__.__name__,
                                       self.vsz, self.rss)

    def update(self):
        """
        Refresh the information using platform instruments. Returns true if
        this operation yields useful values on the current platform.
        """
        return False  # pragma: no cover

    def __sub__(self, other):
        diff = [('Resident set size (delta)', self.rss - other.rss),
                ('Virtual size (delta)', self.vsz - other.vsz),
                ]
        return diff


ProcessMemoryInfo = _ProcessMemoryInfo


def is_available():
    """
    Convenience function to check if the current platform is supported by this
    module.
    """
    return ProcessMemoryInfo().update()


class _ProcessMemoryInfoPS(_ProcessMemoryInfo):
    def update(self):
        """
        Get virtual and resident size of current process via 'ps'.
        This should work for MacOS X, Solaris, Linux. Returns true if it was
        successful.
        """
        try:
            p = Popen(['/bin/ps', '-p%s' % self.pid, '-o', 'rss,vsz'],
                      stdout=PIPE, stderr=PIPE)
        except OSError:  # pragma: no cover
            pass
        else:
            s = p.communicate()[0].split()
            if p.returncode == 0 and len(s) >= 2:  # pragma: no branch
                self.vsz = int(s[-1]) * 1024
                self.rss = int(s[-2]) * 1024
                return True
        return False  # pragma: no cover


class _ProcessMemoryInfoProc(_ProcessMemoryInfo):

    key_map = {
        'VmPeak': 'Peak virtual memory size',
        'VmSize': 'Virtual memory size',
        'VmLck':  'Locked memory size',
        'VmHWM':  'Peak resident set size',
        'VmRSS':  'Resident set size',
        'VmStk':  'Size of stack segment',
        'VmData': 'Size of data segment',
        'VmExe':  'Size of code segment',
        'VmLib':  'Shared library code size',
        'VmPTE':  'Page table entries size',
    }

    def update(self):
        """
        Get virtual size of current process by reading the process' stat file.
        This should work for Linux.
        """
        try:
            stat = open('/proc/self/stat')
            status = open('/proc/self/status')
        except IOError:  # pragma: no cover
            return False
        else:
            stats = stat.read().split()
            self.vsz = int(stats[22])
            self.rss = int(stats[23]) * self.pagesize
            self.pagefaults = int(stats[11])

            for entry in status.readlines():
                try:
                    key, value = entry.split(':', 1)
                except ValueError:
                    continue
                value = value.strip()
                size_in_bytes = lambda x: int(x.split()[0]) * 1024
                if key == 'VmData':
                    self.data_segment = size_in_bytes(value)
                elif key == 'VmExe':
                    self.code_segment = size_in_bytes(value)
                elif key == 'VmLib':
                    self.shared_segment = size_in_bytes(value)
                elif key == 'VmStk':
                    self.stack_segment = size_in_bytes(value)
                key = self.key_map.get(key)
                if key:
                    self.os_specific.append((key, pp(size_in_bytes(value))))

            stat.close()
            status.close()
            return True


try:
    from resource import getrusage, RUSAGE_SELF

    class _ProcessMemoryInfoResource(_ProcessMemoryInfo):
        def update(self):
            """
            Get memory metrics of current process through `getrusage`.  Only
            available on Unix, on Linux most of the fields are not set,
            and on BSD units are used that are not very helpful, see:

            http://www.perlmonks.org/?node_id=626693

            Furthermore, getrusage only provides accumulated statistics (e.g.
            max rss vs current rss).
            """
            usage = getrusage(RUSAGE_SELF)
            self.rss = usage.ru_maxrss * 1024
            self.data_segment = usage.ru_idrss * 1024  # TODO: ticks?
            self.shared_segment = usage.ru_ixrss * 1024  # TODO: ticks?
            self.stack_segment = usage.ru_isrss * 1024  # TODO: ticks?
            self.vsz = (self.data_segment + self.shared_segment +
                        self.stack_segment)

            self.pagefaults = usage.ru_majflt
            return self.rss != 0

    if _ProcessMemoryInfoProc().update():  # pragma: no branch
        ProcessMemoryInfo = _ProcessMemoryInfoProc
    elif _ProcessMemoryInfoPS().update():  # pragma: no cover
        ProcessMemoryInfo = _ProcessMemoryInfoPS
    elif _ProcessMemoryInfoResource().update():  # pragma: no cover
        ProcessMemoryInfo = _ProcessMemoryInfoResource

except ImportError:
    try:
        # Requires pywin32
        from win32process import GetProcessMemoryInfo
        from win32api import GetCurrentProcess, GlobalMemoryStatusEx
    except ImportError:
        logging.warn("Please install pywin32 when using pympler on Windows.")
    else:
        class _ProcessMemoryInfoWin32(_ProcessMemoryInfo):
            def update(self):
                process_handle = GetCurrentProcess()
                meminfo = GetProcessMemoryInfo(process_handle)
                memstatus = GlobalMemoryStatusEx()
                self.vsz = (memstatus['TotalVirtual'] -
                            memstatus['AvailVirtual'])
                self.rss = meminfo['WorkingSetSize']
                self.pagefaults = meminfo['PageFaultCount']
                return True

        ProcessMemoryInfo = _ProcessMemoryInfoWin32


class ThreadInfo(object):
    """Collect information about an active thread."""

    def __init__(self, thread):
        self.ident = thread.ident
        self.name = thread.name
        self.daemon = thread.daemon


def get_current_threads():
    """Get a list of `ThreadInfo` objects."""
    return [ThreadInfo(thread) for thread in threading.enumerate()]


def get_current_thread_id():
    """Get the ID of the current thread."""
    return get_ident()



# https://pandas.pydata.org/pandas-docs/stable/user_guide/10min.html
import pandas as pd
import numpy as np

np.random.seed(0)  # uncomment if you want deterministic outputs
s = pd.Series([1, 3, 5, np.nan, 6, 8])
dates = pd.date_range("20130101", periods=6)
df = pd.DataFrame(np.random.randn(6, 4), index=dates, columns=list("ABCD"))

df.loc["2013-01-03":"2013-01-05", ["C", "B"]]



"""
executes parsed python code to compute the intermediate values within the
last line of the code.
"""

from __future__ import annotations

import builtins
import contextlib
import dataclasses
import sys
import types
import typing as t

import pandas as pd

from . import util
from .parse_nodes import (
    AggCall,
    ChainStatement,
    ChainStep,
    CodeRange,
    EvalError,
    ParseResult,
    ParseSyntaxError,
    PassThroughCall,
    RawCode,
    Subscript,
)

# technically args can be anything...but most of the time it'll be labels
Arg = t.Union[str, list]

Args = t.Dict[str, Arg]


@dataclasses.dataclass
class EvalResult:
    step: ChainStep
    # location of fragment to highlight, relative to the entire expression
    fragment: CodeRange
    args: Args
    val: t.Any


@dataclasses.dataclass
class DFResult(EvalResult):
    val: pd.DataFrame


@dataclasses.dataclass
class SeriesResult(EvalResult):
    val: pd.Series


@dataclasses.dataclass
class GroupbyResult(EvalResult):
    val: util.DataFrameGroupBy


@dataclasses.dataclass
class SeriesGroupbyResult(EvalResult):
    val: util.SeriesGroupBy


@dataclasses.dataclass
class ImageResult(EvalResult):
    val: t.Any


@dataclasses.dataclass
class SyntaxErrorResult(EvalResult):
    step: ParseSyntaxError
    val: None


@dataclasses.dataclass
class RuntimeErrorResult(EvalResult):
    val: Exception


@dataclasses.dataclass
class UnhandledResult(EvalResult):
    """catch-all for chain outputs we don't know how to serialize"""

    val: t.Any


def run(root: ParseResult, ipython_shell=None) -> t.List[EvalResult]:
    """
    runs parsed code. in ipython, we pass in the shell object so we can
    run user code in their namespace.
    """
    # since we send JSON to stdout, we'll send user code's stdout to stderr.
    with contextlib.redirect_stdout(sys.stderr):
        return run_code(root, ipython_shell)


def run_code(root: ParseResult, ipython_shell=None) -> t.List[EvalResult]:
    if isinstance(root, ParseSyntaxError):
        return [
            SyntaxErrorResult(
                step=root, fragment=root.location, args={}, val=None
            )
        ]

    statements = root.statements

    # hard-code the last one!
    setup_stmts, last_expr = statements[:-1], statements[-1]

    # TODO: return warning when last statement isn't a chain
    if not isinstance(last_expr, ChainStatement):
        return []

    # grab user local vars from ipython shell if possible, otherwise initialize
    # a new globals dict
    user_globals = (
        setup_user_globals() if ipython_shell is None else ipython_shell.user_ns
    )

    for stmt in setup_stmts:
        try:
            exec(stmt.code, user_globals)
        except Exception as error:
            step = EvalError.from_node(stmt)
            return [
                RuntimeErrorResult(
                    step=step,
                    fragment=step.location,
                    args={},
                    val=error,
                )
            ]

    relative_to = last_expr.location.start
    last_val: t.Any = None
    eval_results: t.List[EvalResult] = []
    for step in last_expr.chain:
        fragment = step.location % relative_to

        try:
            # wrap individual steps in parens before eval since subexpressions
            # within a line can have newlines
            val = eval(f"({step.code})", user_globals)
            args = eval_args(step, user_globals)
        except Exception as error:
            step = EvalError.from_node(step)
            err_result = RuntimeErrorResult(
                step=step,
                fragment=fragment,
                args={},
                val=error,
            )
            eval_results.append(err_result)
            break

        result = make_result(step, fragment, args, val, last_val)
        eval_results.append(result)
        last_val = val

    # HACK: hard-code all step code strings to the entire statement's code
    # string to make fragment positions work. the code string for individual
    # function calls omits the whitespace that the fragment positions already
    # take into account.
    # https://github.com/SamLau95/pandas_tutor/issues/42
    for result in eval_results:
        result.step.code = last_expr.code
    return eval_results


# need the last_val for the special case where we have an agg that doesn't show
# up immediately after a groupby, like dogs.groupby(...)['...'].mean().
def make_result(
    step: ChainStep,
    fragment: CodeRange,
    args: Args,
    val: t.Any,
    last_val: t.Any,
) -> EvalResult:
    if (
        isinstance(step, PassThroughCall)
        and isinstance(last_val, (util.DataFrameGroupBy, util.SeriesGroupBy))
        and isinstance(val, (pd.DataFrame, pd.Series))
    ):
        step = AggCall.from_passthrough_call(step)

    if isinstance(val, util.DataFrameGroupBy):
        return GroupbyResult(step, fragment, args, val)
    elif isinstance(val, util.SeriesGroupBy):
        return SeriesGroupbyResult(step, fragment, args, val)
    elif isinstance(val, pd.DataFrame):
        return DFResult(step, fragment, args, val)
    elif isinstance(val, pd.Series):
        return SeriesResult(step, fragment, args, val)
    elif util.is_plottable(val):
        return ImageResult(step, fragment, args, val)
    else:
        return UnhandledResult(step, fragment, args, val)


def setup_user_globals():
    # set up scope like PythonTutor
    user_builtins = {}
    assert isinstance(builtins, types.ModuleType)
    for k in dir(builtins):
        user_builtins[k] = getattr(builtins, k)

    user_globals = {}

    user_globals.update({"__name__": "__main__", "__builtins__": user_builtins})

    return user_globals


def eval_args(step: ChainStep, user_globals: dict) -> Args:
    """eval each arg marked with parse_nodes.evals_into()"""
    if isinstance(step, Subscript):
        return eval_args_subscript(step, user_globals)
    return eval_dataclass(step, user_globals)


def eval_args_subscript(step: Subscript, user_globals: dict) -> Args:
    """subscripts have nested eval exprs so we have a special case"""
    slice1_args = eval_dataclass(step.slice1, user_globals, attr="slice1")
    slice2_args = eval_dataclass(step.slice2, user_globals, attr="slice2")
    return {**slice1_args, **slice2_args}


def eval_dataclass(obj: t.Any, user_globals: dict, attr="") -> Args:
    """
    takes a dataclasss with fields marked by evals_into(), outputs
    dict of evaluated values
    """
    args: Args = {}
    if obj is None:
        return args

    fields = [
        field
        for field in dataclasses.fields(obj)
        if field.metadata.get("evals_into", False)
    ]

    for field in fields:
        evals_into = field.metadata["evals_into"].format(attr=attr)
        to_eval: t.Union[None, RawCode, t.List[RawCode]] = getattr(
            obj, field.name
        )
        if to_eval is None:
            continue
        elif isinstance(to_eval, list):
            result = [eval(f"({code})", user_globals) for code in to_eval]
        else:
            result = eval(f"({to_eval})", user_globals)
        args[evals_into] = result
    return args



# Tests that parser can handle different types of susbcript args
df.loc[:, 'Count']

df[['Name', 'Count']]
df[df.columns[:4]]

df[df['Count'] < 10000]

col = 'Name'
df[(df[col] > 10) | (df['Year'] >= 2020)]

mask = df['Count'] > 10
df[mask]

(df
 .iloc[2:5]
 .loc[:, df['tricky'] == 'to parse']
)

(df
 [2:5]
 [df['also'] == 'tricky']
)



# Common point of confusion: difference between loc and iloc
#
# Compare loc_vs_iloc_01.py and loc_vs_iloc_02.py.

import pandas as pd
import io

csv = '''
breed,grooming,food_cost,kids,size
Labrador Retriever,weekly,466.0,high,medium
German Shepherd,weekly,466.0,medium,large
Beagle,daily,324.0,high,small
Golden Retriever,weekly,466.0,high,medium
Yorkshire Terrier,daily,324.0,low,small
Bulldog,weekly,466.0,medium,medium
Boxer,weekly,466.0,high,medium
'''

dogs = pd.read_csv(io.StringIO(csv))

# These return different tables!
dogs.sort_values('kids').iloc[0:6]
# dogs.sort_values('kids').loc[0:6]



import pandas as pd
import io

csv = '''
breed,type,price,longevity,size
Labrador Retriever,sporting,810.0,12.04,medium
German Shepherd,herding,820.0,9.73,large
Beagle,hound,288.0,12.3,small
'''

dogs = pd.read_csv(io.StringIO(csv)).set_index('breed')

(dogs
 .drop(columns=['type', 'price'])
)



# https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.stack.html
import pandas as pd

df_single_level_cols = pd.DataFrame(
    [[0, 1], [2, 3]], index=["cat", "dog"], columns=["weight", "height"]
)

df_single_level_cols.stack()



# https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html
import pandas as pd
import numpy as np

lst = [1, 2, 3, 4, 5, 6] # <-- this was initially [1, 2, 3, 1, 2, 3] so i thought it was a duplicate index issue, but even with no duplicates it still crashes
s = pd.Series([1, 2, 3, 10, 20, 30], lst)
s.groupby(level=0)



import pandas as pd
import io

csv = """
Name,Sex,Count,Year
Noah,M,18252,2020
Julius,M,960,2020
Karen,F,325,2020
"""

csv2 = """
nyt_name,category
Karen,boomer
Julius,mythology
Freya,mythology
"""

baby = pd.read_csv(io.StringIO(csv)).set_index("Name")
nyt = pd.read_csv(io.StringIO(csv2)).set_index("nyt_name")

# tests that output is right when sort=True
baby.join(nyt, sort=True)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
import re
from abc import ABC
from typing import Pattern, Union

import libcst
from libcst.codemod import CodemodContext, VisitorBasedCodemodCommand


class RemovePyreDirectiveCommand(VisitorBasedCodemodCommand, ABC):
    PYRE_TAG: str

    def __init__(self, context: CodemodContext) -> None:
        super().__init__(context)
        self._regex_pattern: Pattern[str] = re.compile(
            rf"^#\s+pyre-{self.PYRE_TAG}\s*$"
        )

    def leave_EmptyLine(
        self, original_node: libcst.EmptyLine, updated_node: libcst.EmptyLine
    ) -> Union[libcst.EmptyLine, libcst.RemovalSentinel]:
        if updated_node.comment is None or not bool(
            self._regex_pattern.search(
                libcst.ensure_type(updated_node.comment, libcst.Comment).value
            )
        ):
            # This is a normal comment
            return updated_node
        # This is a directive comment matching our tag, so remove it.
        return libcst.RemoveFromParent()


class RemovePyreStrictCommand(RemovePyreDirectiveCommand):
    """
    Given a source file, we'll remove the any strict tag if the file already
    contains it.
    """

    DESCRIPTION: str = "Removes the 'pyre-strict' tag from a module."

    PYRE_TAG: str = "strict"


class RemovePyreUnsafeCommand(RemovePyreDirectiveCommand):
    """
    Given a source file, we'll remove the any unsafe tag if the file already
    contains it.
    """

    DESCRIPTION: str = "Removes the 'pyre-unsafe' tag from a module."

    PYRE_TAG: str = "unsafe"



# tests that pointers still trigger memory cap


class MyStr:
    def __init__(self, val):
        self.val = val


# this uses about 2.5 MB
# but, sys.getsizeof() only returns ~120 KB
hello = [MyStr('hello world') for _ in range(15_000)]

hello[:5]



import pandas as pd
import io

csv = """
Name,Sex,Count,Year
Noah,M,18252,2020
Julius,M,960,2020
Karen,M,6,2020
"""

csv2 = """
nyt_name,category
Karen,boomer
Julius,mythology
Freya,mythology
"""

baby = pd.read_csv(io.StringIO(csv))
nyt = pd.read_csv(io.StringIO(csv2))

baby2 = baby.set_index("Name")
nyt2 = nyt.set_index("nyt_name")

# tests index args
baby2.merge(nyt2, left_index=True, right_index=True)



"""marks for join operations, like merge() and join()"""

from typing import List, Optional, Sequence, Tuple, Union, cast

import pandas as pd
from pandas_tutor import util
from pandas_tutor.diagram import AxisPos, Mark, PosPair, Using
from pandas_tutor.marks.mark_utils import (  # noqa: F401
    lhs,
    lhs2,
    lhs2_index,
    lhs2_series,
    lhs_index,
    lhs_series,
    make_drops,
    make_map_sets,
    make_usings,
    print_axis_sets,
    rhs,
    rhs_index,
)
from pandas_tutor.parse_nodes import JoinCall, MergeCall
from pandas_tutor.run import Args, DFResult, EvalResult, SeriesResult
from pandas_tutor.util import Label


def mark_for_merge(
    step: MergeCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    # there's no Series.merge() method, but pd.merge(left, right) can take a
    # series as the left arg
    if not (
        isinstance(before, (DFResult, SeriesResult))
        and isinstance(after, DFResult)
    ):
        return []

    left = before.val
    right = after.val
    args = after.args
    left2 = args.get("right")

    how: str = cast(str, args.get("how", "inner"))
    on_orig = args.get("on")
    left_on_orig = args.get("left_on")
    right_on_orig = args.get("right_on")
    left_index: bool = cast(bool, args.get("left_index", False))
    right_index: bool = cast(bool, args.get("right_index", False))

    # sort=False as default is important to match behavior of pd.merge() since
    # get_join_info() has sort=True
    sort: bool = cast(bool, args.get("sort", False))

    if not util.is_pd(left2):
        return []

    return _merge_marks(
        left,
        left2,
        right,
        how=how,
        on_orig=on_orig,
        left_on_orig=left_on_orig,
        right_on_orig=right_on_orig,
        left_index=left_index,
        right_index=right_index,
        sort=sort,
    )


def mark_for_join(
    step: JoinCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (isinstance(before, DFResult) and isinstance(after, DFResult)):
        return []

    left = before.val
    right = after.val
    args = after.args
    left2 = args.get("other")

    # note that .join() uses a *left join* by default, not inner
    how: str = cast(str, args.get("how", "left"))
    on_orig = args.get("on")

    # sort=False as default is important to match behavior of pd.join() since
    # get_join_info() has sort=True
    sort: bool = cast(bool, args.get("sort", False))

    # .join() can technically take a *list* of dataframes to join with, but we
    # only handle the single dataframe case
    if not util.is_pd(left2):
        return []

    # recreate pandas join() logic from pd.Dataframe._join_compat()

    if how == "cross":
        return _merge_marks(
            left, left2, right, how=how, on_orig=on_orig, sort=sort
        )

    return _merge_marks(
        left,
        left2,
        right,
        how=how,
        left_on_orig=on_orig,
        left_index=on_orig is None,
        right_index=True,
        sort=sort,
    )


def _merge_marks(
    left_orig: Union[pd.DataFrame, pd.Series],
    left2_orig: Union[pd.DataFrame, pd.Series],
    right: pd.DataFrame,
    how: str = "inner",
    on_orig: Union[Label, Sequence[Label], None] = None,
    left_on_orig: Union[Label, Sequence[Label], None] = None,
    right_on_orig: Union[Label, Sequence[Label], None] = None,
    left_index: bool = False,
    right_index: bool = False,
    sort: bool = False,
) -> List[Mark]:
    """
    used to make marks for all merge methods
    """
    left_is_series = isinstance(left_orig, pd.Series)
    left: pd.DataFrame = left_orig.to_frame() if left_is_series else left_orig

    left2_is_series = isinstance(left2_orig, pd.Series)
    left2: pd.DataFrame = (
        left2_orig.to_frame() if left2_is_series else left2_orig
    )

    on = util.listify(on_orig) if on_orig else None
    left_on = (
        on if on_orig else util.listify(left_on_orig) if left_on_orig else None
    )
    right_on = (
        on
        if on_orig
        else util.listify(right_on_orig)
        if right_on_orig
        else None
    )

    if not (on or left_on or right_on or left_index or right_index):
        # default on= is intersection of columns
        on = left_on = right_on = list(left.columns.intersection(left2.columns))

    # don't handle cases where we join using both index and columns since that
    # creates duplicate index labels
    if left_index is not right_index:
        return []

    res_index, left_row_nums, left2_row_nums = util.get_join_info(
        left=left_orig,
        right=left2_orig,
        how=how,
        on=on_orig,
        left_on=left_on_orig,
        right_on=right_on_orig,
        left_index=left_index,
        right_index=right_index,
        sort=sort,
    )
    # a left join on the indexes keeps the same index as the left df, and
    # pandas sets left_row_nums=None.
    if left_row_nums is None:
        left_row_nums = pd.RangeIndex(len(left))
    # same deal for right joins
    if left2_row_nums is None:
        left2_row_nums = pd.RangeIndex(len(left2))

    # mark all columns used for joining
    if left_on and right_on:
        left_on = cast(List, left_on)  # keep mypy happy
        right_on = cast(List, right_on)

        # special case to handle series
        lhs_usings = (
            make_usings(left_on, "column", "lhs")
            if not left_is_series
            else [Using(lhs_series())]
        )

        lhs2_usings = (
            make_usings(right_on, "column", "lhs2")
            if not left2_is_series
            else [Using(lhs2_series())]
        )

        using = [
            *lhs_usings,
            *lhs2_usings,
            *make_usings(on if on else left_on + right_on, "column", "rhs"),
        ]
    else:  # we're joining on the indexes
        using = cast(
            List[Mark],
            [Using(lhs_index("row", i)) for i in range(left.index.nlevels)]
            + [Using(lhs2_index("row", i)) for i in range(left2.index.nlevels)]
            + [Using(rhs_index("row", i)) for i in range(right.index.nlevels)],
        )

    # mark all rows dropped from either lhs or lhs2
    drops = [
        *make_drops(_dropped_labels(left.index, left_row_nums), "row", "lhs"),
        *make_drops(
            _dropped_labels(left2.index, left2_row_nums), "row", "lhs2"
        ),
    ]

    def row_pairs(left_num: int, left2_num: int, right_row: Label):
        left_row = cast(Label, left.index[left_num])
        left2_row = cast(Label, left2.index[left2_num])
        if left_num != -1:
            yield (lhs("row", left_row), rhs("row", right_row))
        if left2_num != -1:
            yield (lhs2("row", left2_row), rhs("row", right_row))
        # don't actually need this last case since if lhs -> rhs and lhs2 ->
        # rhs, we automatically have lhs and lhs2 in mapset together
        # if left_num != -1 and left2_num != -1:
        #     yield (lhs("row", left_row), lhs2("row", left2_row))

    pairs: List[PosPair] = [
        pair
        for left_num, left2_num, right_row in zip(
            left_row_nums, left2_row_nums, res_index
        )
        for pair in row_pairs(left_num, left2_num, right_row)
    ]

    # the merge key is a tuple of row values or an index label from lhs or lhs2
    def by_merge_key(pair: Tuple[AxisPos, AxisPos]):
        # pair is always {lhs, lhs2} -> rhs
        pos, _ = pair
        df = left if pos.anchor == "lhs" else left2
        key = left_on if pos.anchor == "lhs" else right_on

        # BUG: this breaks when the merge keys have index level names!
        # https://github.com/SamLau95/pandas_tutor/issues/154
        row = df.loc[pos.label]
        # if pos.label is duplicated, row is a dataframe
        if isinstance(row, pd.DataFrame):
            row = row.iloc[0]
        return tuple(row.loc[key]) if key else row.name

    row_sets = make_map_sets(pairs, key=by_merge_key)

    # print_axis_sets(row_sets)
    # breakpoint()

    return [*using, *drops, *row_sets]


def _get_left2_arg(args: Args) -> Tuple[Optional[pd.DataFrame], bool]:
    left2_arg = args.get("right")

    # merge with a series treats the series as a 1-column dataframe
    if isinstance(left2_arg, pd.Series):
        return (left2_arg.to_frame(), True)
    return (left2_arg if isinstance(left2_arg, pd.DataFrame) else None, False)


def _dropped_labels(index: pd.Index, row_nums: pd.Index) -> pd.Index:
    dropped = pd.RangeIndex(len(index)).difference(row_nums)
    return index[dropped]



# https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html
import pandas as pd
import numpy as np
np.random.seed(0)

df = pd.DataFrame(
    {
        "A": ["foo", "bar", "foo", "bar", "foo", "bar", "foo", "foo"],
        "B": ["one", "one", "two", "three", "two", "two", "one", "three"],
        "C": np.random.randn(8),
        "D": np.random.randn(8),
    }
)

df2 = df.set_index(["A", "B"])
df2.groupby(level=df2.index.names.difference(["B"])).mean()


import pandas as pd
import io

csv = """
Name,Sex,Count,Year
Noah,M,18252,2020
Julius,M,960,2020
Karen,M,6,2020
Karen,F,325,2020
Noah,F,305,2020
"""

csv2 = """
nyt_name,category
Karen,boomer
Julius,mythology
Freya,mythology
Julius,test
Julius,test2
Karen,boomer2
"""

baby = pd.read_csv(io.StringIO(csv))
nyt = pd.read_csv(io.StringIO(csv2))

nyt2 = nyt.rename(columns={"nyt_name": "Name"})

# Julius and Karen appear multiple times in both lhs and lhs2
baby.merge(nyt2, on="Name")



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import json
import subprocess
from pathlib import Path
from typing import Dict, List, Mapping, Optional, Sequence, Tuple

from mypy_extensions import TypedDict

import libcst as cst
from libcst._position import CodePosition, CodeRange
from libcst.metadata.base_provider import BatchableMetadataProvider
from libcst.metadata.position_provider import PositionProvider


class Position(TypedDict):
    line: int
    column: int


class Location(TypedDict):
    path: str
    start: Position
    stop: Position


class InferredType(TypedDict):
    location: Location
    annotation: str


class PyreData(TypedDict, total=False):
    types: Sequence[InferredType]


class TypeInferenceProvider(BatchableMetadataProvider[str]):
    """
    Access inferred type annotation through `Pyre Query API <https://pyre-check.org/docs/querying-pyre.html>`_.
    It requires `setup watchman <https://pyre-check.org/docs/watchman-integration.html>`_
    and start pyre server by running ``pyre`` command.
    The inferred type is a string of `type annotation <https://docs.python.org/3/library/typing.html>`_.
    E.g. ``typing.List[libcst._nodes.expression.Name]``
    is the inferred type of name ``n`` in expression ``n = [cst.Name("")]``.
    All name references use the fully qualified name regardless how the names are imported.
    (e.g. ``import libcst; libcst.Name`` and ``import libcst as cst; cst.Name`` refer to the same name.)
    Pyre infers the type of :class:`~libcst.Name`, :class:`~libcst.Attribute` and :class:`~libcst.Call` nodes.
    The inter process communication to Pyre server is managed by :class:`~libcst.metadata.FullRepoManager`.
    """

    METADATA_DEPENDENCIES = (PositionProvider,)

    @staticmethod
    def gen_cache(
        root_path: Path, paths: List[str], timeout: Optional[int]
    ) -> Mapping[str, object]:
        params = ",".join(f"path='{root_path / path}'" for path in paths)
        cmd_args = ["pyre", "--noninteractive", "query", f"types({params})"]
        try:
            stdout, stderr, return_code = run_command(cmd_args, timeout=timeout)
        except subprocess.TimeoutExpired as exc:

            raise exc

        if return_code != 0:
            raise Exception(f"stderr:\n {stderr}\nstdout:\n {stdout}")
        try:
            resp = json.loads(stdout)["response"]
        except Exception as e:
            raise Exception(f"{e}\n\nstderr:\n {stderr}\nstdout:\n {stdout}")
        return {path: _process_pyre_data(data) for path, data in zip(paths, resp)}

    def __init__(self, cache: PyreData) -> None:
        super().__init__(cache)
        lookup: Dict[CodeRange, str] = {}
        cache_types = cache.get("types", [])
        for item in cache_types:
            location = item["location"]
            start = location["start"]
            end = location["stop"]
            lookup[
                CodeRange(
                    start=CodePosition(start["line"], start["column"]),
                    end=CodePosition(end["line"], end["column"]),
                )
            ] = item["annotation"]
        self.lookup: Dict[CodeRange, str] = lookup

    def _parse_metadata(self, node: cst.CSTNode) -> None:
        range = self.get_metadata(PositionProvider, node)
        if range in self.lookup:
            self.set_metadata(node, self.lookup.pop(range))

    def visit_Name(self, node: cst.Name) -> Optional[bool]:
        self._parse_metadata(node)

    def visit_Attribute(self, node: cst.Attribute) -> Optional[bool]:
        self._parse_metadata(node)

    def visit_Call(self, node: cst.Call) -> Optional[bool]:
        self._parse_metadata(node)


def run_command(
    cmd_args: List[str], timeout: Optional[int] = None
) -> Tuple[str, str, int]:
    process = subprocess.run(cmd_args, capture_output=True, timeout=timeout)
    return process.stdout.decode(), process.stderr.decode(), process.returncode


class RawPyreData(TypedDict):
    path: str
    types: Sequence[InferredType]


def _process_pyre_data(data: RawPyreData) -> PyreData:
    return {"types": sorted(data["types"], key=_sort_by_position)}


def _sort_by_position(data: InferredType) -> Tuple[int, int, int, int]:
    start = data["location"]["start"]
    stop = data["location"]["stop"]
    return start["line"], start["column"], stop["line"], stop["column"]



"""
has dataclass definitions for final JSON output.
"""

from __future__ import annotations

import dataclasses
from dataclasses import field
from traceback import TracebackException
from typing import Any, List, Tuple, Union

import pandas as pd
import simplejson as json
from typing_extensions import Literal

from .parse_nodes import ParseSyntaxError
from .run import RuntimeErrorResult
from .util import (
    SERIES,
    CodePosition,
    CodeRange,
    JSONScalar,
    Label,
    df_data,
    index_data,
    json_scalar,
    series_data,
    unwrap,
)


@dataclasses.dataclass
class OutputSpec:
    """the final object we'll make into JSON"""

    code: str
    explanation: Explanation

    def to_json(self):
        return json.dumps(
            self, indent=2, default=encode_dataclasses, ignore_nan=True
        )


def _diagram_as_dict(dclass):
    """pass into dataclasses.asdict to rename from_ to from"""
    res = dict(dclass)
    # we want to preserve the original dict order, so we rebuild the dict if we
    # see from_
    if "from_" in res:
        return {"from" if k == "from_" else k: v for k, v in res.items()}
    return res


def encode_dataclasses(obj: Any):
    if dataclasses.is_dataclass(obj):
        return dataclasses.asdict(obj, dict_factory=_diagram_as_dict)
    return json_scalar(obj)


##############################################################################
# Explanation
##############################################################################


@dataclasses.dataclass
class Diagram:
    type: str
    code_step: str
    fragment: CodeRange
    marks: List[Mark]
    data: Union[DataPair, DataTwoLHS]


@dataclasses.dataclass
class ErrorOutput:
    type: str = field(default="ErrorOutput", init=False, repr=False)
    code_step: str
    message: str


@dataclasses.dataclass
class SyntaxErrorOutput(ErrorOutput):
    type: str = field(default="SyntaxErrorOutput", init=False, repr=False)
    location: CodePosition

    @classmethod
    def from_parse_syntax_error(cls, err: ParseSyntaxError):
        return cls(
            code_step=err.code,
            message=err.error_msg,
            location=err.location.start,
        )


@dataclasses.dataclass
class RuntimeErrorOutput(ErrorOutput):
    type: str = field(default="RuntimeErrorOutput", init=False, repr=False)
    fragment: CodeRange

    @classmethod
    def from_runtime_error_result(cls, result: RuntimeErrorResult):
        tb = TracebackException.from_exception(result.val)
        # get error message from last stack frame
        message = list(tb.format_exception_only())[-1]
        return cls(
            code_step=result.step.code,
            message=message,
            fragment=result.fragment,
        )


@dataclasses.dataclass
class RuntimeErrorInSetup(RuntimeErrorOutput):
    type: str = field(default="RuntimeErrorInSetup", init=False, repr=False)


@dataclasses.dataclass
class RuntimeErrorInChain(RuntimeErrorOutput):
    type: str = field(default="RuntimeErrorInChain", init=False, repr=False)


Explanation = List[Union[Diagram, ErrorOutput]]

##############################################################################
# TablePos
##############################################################################

# the table we're pointing to
Anchor = Literal["lhs", "rhs", "lhs2"]

# the axis we're pointing to
Selection = Literal["column", "row"]

# the index level we're pointing to. None if index is not multi-level
IndexLevel = Union[None, int]

# each TablePos object is serialized as one of these
TablePosType = Literal["axis", "series", "index_level", "cell"]


@dataclasses.dataclass
class TablePos:
    """
    base class that represents a position in a table or series.
    we use this to point to:

    - an single column or row
    - a entire series
    - a single label in the row or column index
    - an entire level in the column or row index
    - the name for an index level
    - a single datum
    """

    # needs to be initialized by subclass
    type: TablePosType = field(init=False)

    def __post_init__(self):
        raise NotImplementedError("subclasses need to initialize self.type")


PosPair = Tuple[TablePos, TablePos]


@dataclasses.dataclass
class AxisPos(TablePos):
    """points to a single column or row for a table"""

    anchor: Anchor
    select: Selection
    label: Label

    def __post_init__(self):
        self.type = "axis"


@dataclasses.dataclass
class SeriesPos(TablePos):
    """points to an entire series"""

    anchor: Anchor
    label: Label = field(default=SERIES, init=False)

    def __post_init__(self):
        self.type = "series"


@dataclasses.dataclass
class IndexLevelPos(TablePos):
    """points to a single level for the index of the column or row labels"""

    anchor: Anchor
    select: Selection
    level: IndexLevel = None

    def __post_init__(self):
        self.type = "index_level"


@dataclasses.dataclass
class CellPos(TablePos):
    """
    points to a single cell in the table, which is uniquely identified by
    the combination of column and row label.
    """

    anchor: Anchor
    row: Label
    # for Series, column is util.SERIES
    column: Label

    def __post_init__(self):
        self.type = "cell"
        self.row = unwrap(self.row)
        self.column = unwrap(self.column)


##############################################################################
# Mark
##############################################################################

# each Mark object is serialized as one of these
MarkType = Literal["using", "map", "map_set", "drop"]


@dataclasses.dataclass
class Mark:
    """base class, don't use directly"""

    type: MarkType = field(init=False)

    def __post_init__(self):
        raise NotImplementedError(
            "subclasses need to initialize self.illustrate"
        )


@dataclasses.dataclass
class Using(Mark):
    """represents the data we used to perform an operation"""

    pos: TablePos

    def __post_init__(self):
        self.type = "using"


@dataclasses.dataclass
class Map(Mark):
    """represents data copied or mapped from lhs to rhs"""

    # from is a Python keyword!
    from_: TablePos
    to: TablePos

    def __post_init__(self):
        self.type = "map"


@dataclasses.dataclass
class MapSet(Mark):
    """
    represents a set of Map objects. the frontend uses this to simplify the
    visualization when there are many small marks to visualize (e.g. many
    cell-to-cell mappings).
    """

    maps: List[Map]

    def __post_init__(self):
        self.type = "map_set"


@dataclasses.dataclass
class Drop(Mark):
    """represents data explicitly removed from the table"""

    pos: TablePos

    def __post_init__(self):
        self.type = "drop"


##############################################################################
# DataPair and DataFrames
##############################################################################

PrevRHS = Literal["prev_rhs"]
NoRHS = Literal["no_rhs"]


@dataclasses.dataclass
class DataPair:
    lhs: Union[DataSpec, PrevRHS]
    rhs: Union[DataSpec, NoRHS]


@dataclasses.dataclass
class DataTwoLHS:
    """two lhs tables, used for joins"""

    lhs: Union[DataSpec, PrevRHS]
    lhs2: DataSpec
    rhs: DataSpec


@dataclasses.dataclass
class DataSpec:
    """base class for a python val we're going to serialize"""

    type: str


@dataclasses.dataclass
class DFSpec(DataSpec):
    type: str = field(default="DataFrame", init=False, repr=False)
    columns: Index
    index: Index
    data: List[List[JSONScalar]]

    @classmethod
    def from_pd(cls, df: pd.DataFrame) -> DFSpec:
        return cls(
            columns=Index.from_pd(df.columns),
            index=Index.from_pd(df.index),
            data=df_data(df),
        )


@dataclasses.dataclass
class SeriesSpec(DataSpec):
    type: str = field(default="Series", init=False, repr=False)
    index: Index
    data: List[JSONScalar]

    @classmethod
    def from_pd(cls, series: pd.Series) -> SeriesSpec:
        return cls(
            index=Index.from_pd(series.index),
            data=series_data(series),
        )


@dataclasses.dataclass
class Index:
    """represents a pandas index in the serialized data"""

    # names of each index level. unnamed levels are None
    names: Tuple

    # for a multi-index, the labels are a list of tuples. this matches the
    # behavior of pd.Index.tolist()
    labels: List[JSONScalar]

    @classmethod
    def from_pd(cls, index: pd.Index) -> Index:
        return cls(names=tuple(index.names), labels=index_data(index))


@dataclasses.dataclass
class GroupBySpec(DFSpec):
    type: str = field(default="DataFrameGroupBy", init=False, repr=False)
    group_data: GroupData


@dataclasses.dataclass
class SeriesGroupBySpec(SeriesSpec):
    type: str = field(default="SeriesGroupBy", init=False, repr=False)
    group_data: GroupData


@dataclasses.dataclass
class GroupData:
    # grouping cols, if we can pull them out
    columns: List[Label]
    groups: List[Group]


@dataclasses.dataclass
class Group:
    """a group maps between dataframe values -> labels that match"""

    # the group name is the unique combo of grouping vals, so if we do:
    # >>> dogs.groupby(['size', 'kids'])
    # then the group names will be: ['small', 'low'], ['small', 'high'], ...
    #
    # the labels appear in the same order as GroupData.col_names
    name: list

    # labels for all rows the group contains
    labels: List[Label]


@dataclasses.dataclass
class ImageSpec(DataSpec):
    """encodes an image as a base64 png"""

    type: str = field(default="Image", init=False, repr=False)
    data: str


@dataclasses.dataclass
class UnhandledData(DataSpec):
    """catch-all for data that we don't know how to handle, like scalars"""

    type: str = field(default="Unhandled", init=False, repr=False)
    data: JSONScalar



# in Jupyter, load the %%pandas_tutor magic.
# users have to initialize the magic first using %load_ext pandas_tutor.
def load_ipython_extension(ipython):
    from importlib import reload
    import pandas_tutor.ipython_magics

    reload(pandas_tutor.ipython_magics)
    ipython.register_magics(pandas_tutor.ipython_magics.PandasTutorMagics)



"""
creates mark specs. here's where the magic happens!
"""
from typing import (
    Any,
    Callable,
    List,
    Optional,
    Tuple,
    Union,
    cast,
)

import pandas as pd

from .merge import mark_for_join, mark_for_merge

from .mark_utils import (
    diff_rows,
    diff_cols,
    no_marks,
    selection,
    make_usings,
    make_maps,
    make_drops,
    using_and_map,
    lhs,
    rhs,
    lhs_index,
    rhs_index,
    rhs_series,
    by_row,
    by_column,
    by_result_cell,
    make_map_sets,
)

from pandas_tutor import util
from pandas_tutor.diagram import (
    CellPos,
    Drop,
    IndexLevelPos,
    Map,
    Mark,
    PosPair,
    Selection,
)
from pandas_tutor.parse_nodes import (
    AggCall,
    ApplyCall,
    AssignCall,
    ChainStep,
    DropCall,
    EvalError,
    GroupByCall,
    HeadCall,
    JoinCall,
    MeltCall,
    MergeCall,
    PassThroughCall,
    PivotCall,
    PivotTableCall,
    RenameCall,
    ResetIndexCall,
    SetIndexCall,
    SortValuesCall,
    StackCall,
    SubsComparison,
    Subscript,
    SubscriptEl,
    TailCall,
    UnstackCall,
)
from pandas_tutor.run import (
    Arg,
    DFResult,
    EvalResult,
    GroupbyResult,
    SeriesGroupbyResult,
    SeriesResult,
)
from pandas_tutor.util import SERIES, Label, LabelPair


# step comes from after.step, but we pull it out here to help with
# the type checker.
def make_marks(
    step: ChainStep, before: EvalResult, after: EvalResult
) -> List[Mark]:
    """
    computes the marks for a given step by dispatching to the right marks
    function. returns empty list if we don't know how to make marks.
    """
    if isinstance(step, EvalError):
        return no_marks()
    elif isinstance(step, PassThroughCall):
        return no_marks()
    elif isinstance(step, SortValuesCall):
        return mark_for_sort_values(step, before, after)
    elif isinstance(step, DropCall):
        return mark_for_drop(step, before, after)
    elif isinstance(step, RenameCall):
        return mark_for_rename(step, before, after)
    elif isinstance(step, HeadCall) or isinstance(step, TailCall):
        return mark_for_head_or_tail(step, before, after)
    elif isinstance(step, ApplyCall):
        return mark_for_apply(step, before, after)
    elif isinstance(step, AssignCall):
        return mark_for_assign(step, before, after)
    elif isinstance(step, GroupByCall):
        return mark_for_groupby(step, before, after)
    elif isinstance(step, AggCall):
        return mark_for_agg(step, before, after)
    elif isinstance(step, ResetIndexCall):
        return mark_for_reset_index(step, before, after)
    elif isinstance(step, SetIndexCall):
        return mark_for_set_index(step, before, after)
    elif isinstance(step, UnstackCall):
        return mark_for_unstack(step, before, after)
    elif isinstance(step, StackCall):
        return mark_for_stack(step, before, after)
    elif isinstance(step, PivotCall):
        return mark_for_pivot(step, before, after)
    elif isinstance(step, PivotTableCall):
        return mark_for_pivot_table(step, before, after)
    elif isinstance(step, MeltCall):
        return mark_for_melt(step, before, after)
    elif isinstance(step, MergeCall):
        return mark_for_merge(step, before, after)
    elif isinstance(step, JoinCall):
        return mark_for_join(step, before, after)
    elif isinstance(step, Subscript):
        return mark_for_subscript(step, before, after)
    else:
        return no_marks()


# df.sort_values('Name')
def mark_for_sort_values(
    step: SortValuesCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (
        isinstance(before, (DFResult, SeriesResult))
        and isinstance(after, (DFResult, SeriesResult))
    ):
        return []
    df = before.val
    args = after.args

    sort_by = args.get("labels", [])
    if isinstance(sort_by, str):
        sort_by = [sort_by]

    sorted_labels = df.index if step.axis == "index" else df.columns

    # highlight sorted cols in RHS since the LHS values aren't sorted
    highlights = make_usings(
        sort_by, selection(step.axis, other=True), anchor="rhs"
    )
    outlines = make_maps(sorted_labels, selection(step.axis))
    return [*highlights, *outlines]


# dogs.drop(columns=['type', 'price'])
def mark_for_drop(
    step: DropCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (
        isinstance(before, (DFResult, SeriesResult))
        and isinstance(after, (DFResult, SeriesResult))
    ):
        return []
    args = after.args

    col_labels = args.get("col_labels", [])
    if not util.is_list_like(col_labels):
        col_labels = [col_labels]

    row_labels = args.get("row_labels", [])
    if not util.is_list_like(row_labels):
        row_labels = [row_labels]

    # cross out dropped rows or columns
    return [
        *make_drops(col_labels, "column"),
        *make_drops(row_labels, "row"),
    ]


# df.rename(index={'sam': 'smae'})
def mark_for_rename(
    step: RenameCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    args = after.args
    mapping: Any = args.get("mapping", {})

    if not isinstance(mapping, dict):
        return no_marks()

    select = selection(step.axis)

    return [
        Map(from_=lhs(select, old), to=rhs(select, new))
        for old, new in mapping.items()
    ]


# df.head(2)
# df.tail()
def mark_for_head_or_tail(
    step: Union[HeadCall, TailCall], before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (
        isinstance(before, (DFResult, SeriesResult))
        and isinstance(after, (DFResult, SeriesResult))
    ):
        return []
    return diff_rows(before.val, after.val)


# df['breed'].apply(len)
def mark_for_apply(
    step: ApplyCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if isinstance(before, DFResult) and isinstance(after, DFResult):
        df = after.val
        labels = df.index if step.axis == "index" else df.columns
        return make_maps(labels, selection(step.axis))
    elif isinstance(before, DFResult) and isinstance(after, SeriesResult):
        # dogs.apply(len)  -> series with column names as index
        # special case: result is transposed! but i think this is confusing to
        # draw arrows for (since we draw arrows from column to rows) so let's
        # not bother with this.
        return []
    elif isinstance(before, SeriesResult) and isinstance(after, SeriesResult):
        labels = after.val.index
        return make_maps(labels, "row")
    else:
        # TODO: handle apply on groupby objects
        return []


# don't do anything super crazy for assigns...just highlight the new columns
# dogs.assign(daily=lambda df: df['food_cost'] * 30)
def mark_for_assign(
    step: AssignCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    return make_usings(step.new_col_labels, "column", "rhs")


# df.groupby('hello')
def mark_for_groupby(
    step: GroupByCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not isinstance(after, GroupbyResult):
        return []

    df = before.val

    # if user specifies manual groups, the groups don't come from a column in
    # the original table
    group_cols = [
        label
        for label in util.grouping_labels(after.val)
        if label in df.columns
    ]
    highlights = make_usings(
        group_cols, selection(step.axis, other=True), anchor="lhs"
    )

    return highlights


# basic heuristic: assume that group keys map to row labels of result
def mark_for_agg(
    step: AggCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not isinstance(before, (GroupbyResult, SeriesGroupbyResult)):
        return []
    if not isinstance(after, (DFResult, SeriesResult)):
        return []

    groups = util.get_groups(before.val)

    row_outlines: List[Mark] = []
    for group_key, lhs_labels in groups.items():
        for label in lhs_labels:
            row_outlines.append(
                # TODO: get selection from groupby instead of hard-coding 'row'
                Map(
                    from_=lhs("row", label),
                    to=rhs("row", group_key),
                )
            )

    return row_outlines


# dogs.reset_index(level=[1, 2], drop=True)
# i don't think this works properly when the column is a multi-index, but
# let's not worry about that for now
def mark_for_reset_index(
    step: ResetIndexCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (
        isinstance(before, (DFResult, SeriesResult))
        and isinstance(after, DFResult)
    ):
        return []
    df = before.val
    args = after.args

    # if level unspecified, pandas resets all levels
    all_levels = list(range(len(df.index.names)))
    levels = util.listify(args.get("level", all_levels))
    levels = [util.level_number(df.index, level) for level in levels]

    if args.get("drop", False):
        return [Drop(IndexLevelPos("lhs", "row", level)) for level in levels]

    # recreating the pandas defaults for unnamed index levels
    names: List[str]
    if util.is_multi(df.index):
        names = [
            name if name is not None else f"level_{position}"
            for position, name in enumerate(df.index.names)
        ]
    else:
        default = "index" if "index" not in df else "level_0"
        names = [default] if df.index.name is None else [df.index.name]

    return [
        Map(
            IndexLevelPos("lhs", "row", level),
            rhs("column", names[level]),
        )
        for level in levels
    ]


def mark_for_set_index(
    step: SetIndexCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (isinstance(before, DFResult) and isinstance(after, DFResult)):
        return []
    df = before.val
    args = after.args

    keys = util.listify(args.get("keys", []))
    append = args.get("append", False)

    # if append=True, pandas appends new index levels after old levels
    n_orig_levels = len(df.index.names) if append else 0

    return [
        mark
        for index_level, level in enumerate(keys)
        for mark in using_and_map(
            lhs("column", level),
            rhs_index("row", index_level + n_orig_levels),
        )
    ]


# counts.unstack(level=-1, fill_value=0)
def mark_for_unstack(
    step: UnstackCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (
        isinstance(before, (DFResult, SeriesResult))
        and isinstance(after, DFResult)
    ):
        return []

    df = before.val
    args = after.args

    # normally, pandas unstacks the index into the columns. but when there's
    # only one index level, pandas instead transposes the dataframe, then
    # *stacks* a level. it's a pretty strange edge case to draw arrows for, so
    # we don't handle it.
    if not util.is_multi(df.index):
        return []

    levels = util.listify(args.get("level", -1))
    levels = [util.level_number(df.index, level) for level in levels]

    columns = df.columns if util.is_dataframe(df) else pd.Index([util.SERIES])
    n_orig_levels = len(columns.names) if util.is_dataframe(df) else 0

    index_marks: List[Mark] = [
        mark
        for index_level, level in enumerate(levels)
        for mark in using_and_map(
            lhs_index("row", level),
            # unstacking puts the new levels **under** the existing ones
            rhs_index("column", index_level + n_orig_levels),
        )
    ]

    # for each cell: the unstacked labels move to the column index
    cells = util.push_levels(df.index, columns, levels)
    pairs: List[PosPair] = [
        (CellPos("lhs", old_row, old_col), CellPos("rhs", new_row, new_col))
        for (old_row, old_col), (new_row, new_col) in cells
    ]
    # group together marks that map the same column
    cell_sets = make_map_sets(pairs, key=by_column)

    return [*index_marks, *cell_sets]


# counts.stack(level=-1, drop_na=False)
def mark_for_stack(
    step: StackCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (
        isinstance(before, (DFResult))
        and isinstance(after, (DFResult, SeriesResult))
    ):
        return []

    df = before.val
    args = after.args

    levels = util.listify(args.get("level", len(df.columns.names) - 1))
    levels = [util.level_number(df.columns, level) for level in levels]

    n_index_levels = len(df.index.names)

    index_marks: List[Mark] = [
        mark
        for index_level, level in enumerate(levels)
        for mark in using_and_map(
            IndexLevelPos("lhs", "column", level),
            # stacking puts the new levels **after** the existing ones
            IndexLevelPos("rhs", "row", index_level + n_index_levels),
        )
    ]

    # for each cell: the unstacked labels move to the row index
    is_series = isinstance(after, SeriesResult)
    cells = util.push_levels(df.columns, df.index, levels)
    pairs: List[PosPair] = [
        (
            CellPos("lhs", old_row, old_col),
            CellPos("rhs", new_row, new_col if not is_series else util.SERIES),
        )
        for (old_col, old_row), (new_col, new_row) in cells
    ]
    # group together marks that map the same row
    cell_sets = make_map_sets(pairs, key=by_row)

    return [*index_marks, *cell_sets]


# df.pivot(index='foo', columns='bar', values='baz')
def mark_for_pivot(
    step: PivotCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    # if index=[], pandas does the weird transpose + stack thing into a series
    # which we won't try to handle
    if not (isinstance(before, DFResult) and isinstance(after, DFResult)):
        return []

    df = before.val
    args = after.args

    has_index = "index" in args
    has_values = "values" in args
    index: List[Label] = util.listify(args.get("index", []))
    columns: List[Label] = util.listify(args.get("columns", []))
    # default values arg is all leftover columns
    values: List[Label] = util.listify(
        args.get("values", df.columns.drop([*index, *columns]))
    )
    no_value_cols = len(values) == 0

    # special case: when only one values column is specified, pandas
    # doesn't keep it as a column. we need has_values since pandas only drops
    # when values is explicitly passed in.
    will_drop_values = has_values and len(values) == 1
    n_orig_col_levels = len(df.columns.names) if not will_drop_values else 0

    index_marks = [
        mark
        for position, name in enumerate(index)
        for mark in using_and_map(
            lhs("column", name), rhs_index("row", position)
        )
    ]

    # special case: result is empty dataframe with new index
    if no_value_cols:
        return [*index_marks]

    # each column arg is appended as an index level into the columns
    column_marks = [
        mark
        for position, name in enumerate(columns)
        for mark in using_and_map(
            lhs("column", name),
            rhs_index("column", position + n_orig_col_levels),
        )
    ]

    # to make cell marks, we need to pull row and column labels from the data
    # rows themselves so the logic is tricky
    pairs = []
    for old_row, row in df.iterrows():
        old_row = cast(Label, old_row)
        # pull new row labels from row data
        new_row = cast(Label, tuple(row[index]) if has_index else old_row)

        # pull new col labels from row data
        appended = tuple(row[columns])
        for old_col in values:
            new_col = (old_col, *appended) if not will_drop_values else appended
            left = CellPos("lhs", old_row, old_col)
            right = CellPos("rhs", new_row, new_col)
            pairs.append((left, right))

    # group together marks using their original columns
    cell_sets = make_map_sets(pairs, key=by_column)

    return [*index_marks, *column_marks, *cell_sets]


# df.pivot(index='foo', columns='bar', values='baz')
def mark_for_pivot_table(
    step: PivotTableCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    # if index=[], pandas does the weird transpose + stack thing into a series
    # which we won't try to handle
    if not (isinstance(before, DFResult) and isinstance(after, DFResult)):
        return []

    df = before.val
    after_df = after.val
    args = after.args

    has_index = "index" in args
    has_values = "values" in args
    index: List[Label] = util.listify(args.get("index", []))
    columns: List[Label] = util.listify(args.get("columns", []))
    # default values arg is all leftover columns
    values: List[Label] = util.listify(
        args.get("values", df.columns.drop([*index, *columns]))
    )
    aggfunc: Union[str, Callable, list, dict] = args.get("aggfunc", "mean")

    # when multiple aggfuncs are specified, pandas puts the aggfuncs into
    # another level of the column index.
    has_multi_aggs = isinstance(aggfunc, (list, tuple)) or (
        isinstance(aggfunc, dict)
        and any(isinstance(val, (list, tuple)) for val in aggfunc.values())
    )

    # special case: when only one values column is specified, pandas
    # doesn't keep it as a column. we need has_values since pandas only drops
    # when values is explicitly passed in. also drop values when only column
    # levels passed in
    will_drop_values = (has_values and len(values) == 1) or not has_index
    no_value_cols = len(values) == 0
    n_orig_col_levels = len(df.columns.names) if not will_drop_values else 0

    # each index arg goes into a new index level
    index_marks = [
        mark
        for position, name in enumerate(index)
        for mark in using_and_map(
            lhs("column", name), rhs_index("row", position)
        )
    ]

    # special case: result is empty dataframe with new index
    if no_value_cols:
        return [*index_marks]

    # each column arg is appended as an index level into the columns
    column_marks: List[Mark] = [
        mark
        for position, name in enumerate(columns)
        for mark in using_and_map(
            lhs("column", name),
            rhs_index("column", position + n_orig_col_levels),
        )
    ]

    # don't handle cases with multiple agg funcs since the logic is complicated
    if has_multi_aggs:
        return [*index_marks, *column_marks]

    # internally, pandas uses a groupby + unstack to pivot so we'll follow
    # similar logic
    keys = [
        label
        for label in index + columns
        if isinstance(label, str) and label in df.columns
    ]
    column_levels = list(range(len(index), len(keys)))
    groups = util.get_groups(df.groupby(keys))

    def unstack_group(labels, old_col) -> LabelPair:
        return (
            util.push_level(
                labels,
                old_col if not will_drop_values else SERIES,
                column_levels,
            )
            if has_index
            # if no index arg, there's only the column arg. pandas groups using
            # the column arg, then *transposes* the result.
            else (old_col, labels)
        )

    label_pairs: List[Tuple[LabelPair, LabelPair]] = [
        ((old_row, old_col), unstack_group(labels, old_col))
        for labels, old_rows in groups.items()
        for old_row in old_rows
        for old_col in values
    ]
    # take out cells that didn't get agg'd
    pairs: List[PosPair] = [
        (CellPos("lhs", old_row, old_col), CellPos("rhs", new_row, new_col))
        for ((old_row, old_col), (new_row, new_col)) in label_pairs
        if new_col in after_df and new_row in after_df.index
    ]

    # mapset for pivot_table() is more granular than pivot() since we want
    # to show each individual aggregation
    cell_sets = make_map_sets(pairs, key=by_result_cell)

    return [*index_marks, *column_marks, *cell_sets]


def mark_for_melt(
    step: MeltCall, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if not (isinstance(before, DFResult) and isinstance(after, DFResult)):
        return []
    df = before.val
    args = after.args

    # don't handle multi-index melt since it adds a lot of complexity
    if util.is_multi(df.columns):
        return []

    id_vars: List[Label] = util.listify(args.get("id_vars", []))
    # default values arg is all leftover columns
    value_vars: List[Label] = util.listify(
        args.get("value_vars", df.columns.drop(id_vars))
    )
    var_name = cast(
        str,
        args.get(
            "var_name",
            df.columns.name if df.columns.name is not None else "variable",
        ),
    )
    value_name = cast(str, args.get("value_name", "value"))
    ignore_index = args.get("ignore_index", True)

    # multi-index melt adds a lot of complexity, and ignore_index=False
    # duplicates index labels so we don't handle it
    if util.is_multi(df.columns) or not ignore_index:
        return []

    pairs = []
    for (row_num, row) in enumerate(df.index):
        for (col_num, col) in enumerate(value_vars):
            new_row = len(df) * col_num + row_num
            pairs.append(
                (CellPos("lhs", row, col), CellPos("rhs", new_row, var_name))
            )
            pairs.append(
                (CellPos("lhs", row, col), CellPos("rhs", new_row, value_name))
            )

    return make_map_sets(pairs, key=by_column)


# handler for all subscripts, like:
#
# df.loc[1:5, ['Name', 'Count']]
# df.iloc[2:5, 1:4]
# df[df['Count'] > 10000]
# df.groupby('Sex')[['Count']]
def mark_for_subscript(
    step: Subscript, before: EvalResult, after: EvalResult
) -> List[Mark]:
    if isinstance(before, (DFResult, GroupbyResult)) and isinstance(
        after, (SeriesResult, SeriesGroupbyResult)
    ):
        return mark_for_subscript_into_series(step, before, after)
    elif isinstance(before, SeriesResult) and isinstance(after, SeriesResult):
        return mark_for_subscript_of_series(step, before, after)
    elif isinstance(before, (DFResult, GroupbyResult)) and isinstance(
        after, (DFResult, GroupbyResult)
    ):
        return mark_for_subscript_df_to_df(step, before, after)
    else:
        return []


def mark_for_subscript_df_to_df(
    step: Subscript,
    before: Union[DFResult, GroupbyResult],
    after: Union[DFResult, GroupbyResult],
) -> List[Mark]:
    row_slice = step.slice1
    col_slice = step.slice2
    args = after.args

    before_df = util.ungroup(before.val)
    after_df = util.ungroup(after.val)

    # df.loc[:, df.iloc[0] % 2 == 0]
    rows_for_filter = make_subscript_comparison_marks(
        col_slice, args.get("slice2_filter_labels", []), "row"
    )

    no_filter_rows = len(rows_for_filter) == 0

    # df[df['Count'] > 14000]
    cols_for_filter = make_subscript_comparison_marks(
        row_slice, args.get("slice1_filter_labels", []), "column"
    )

    no_filter_cols = len(cols_for_filter) == 0

    return [
        *cols_for_filter,
        # if we're filtering, always display arrows between matching rows/cols
        *diff_cols(before_df, after_df, only_if_diff=no_filter_rows),
        *rows_for_filter,
        *diff_rows(before_df, after_df, only_if_diff=no_filter_cols),
    ]


def make_subscript_comparison_marks(
    subs_el: Optional[SubscriptEl],
    labels: Arg,
    selection: Selection,
) -> List[Mark]:
    """
    makes highlights for cols/rows used for filtering, if the subscript is a
    filter.
    """
    return (
        make_usings(labels, selection)
        if isinstance(subs_el, SubsComparison)
        else []
    )


def mark_for_subscript_of_series(
    step: Subscript,
    before: SeriesResult,
    after: SeriesResult,
) -> List[Mark]:
    # no special cases for comparisons since there isn't a "column" we're using
    # to filter
    return diff_rows(before.val, after.val)


def mark_for_subscript_into_series(
    step: Subscript,
    before: Union[DFResult, GroupbyResult],
    after: Union[SeriesResult, SeriesGroupbyResult],
) -> List[Mark]:
    args = after.args
    before_df = util.ungroup(before.val)
    after_df = util.ungroup(after.val)

    # df['kids']
    if step.slicer is None:
        col = args.get("slice1_values")
        if not isinstance(col, (str, int)):
            return []

        return [Map(from_=lhs("column", col), to=rhs_series())]

    # df.loc[df["email"] > "s", "web"]
    # df.loc[:, df.iloc[0] % 2 == 0]
    row_slice = step.slice1
    col_slice = step.slice2

    # df.loc['sam@sam.com', df.loc['jan@jan.com'] > 10]
    rows_used_for_filter = make_subscript_comparison_marks(
        col_slice, args.get("slice2_filter_labels", []), "row"
    )

    # df.loc[df['Count'] > 14000, 'Name']
    cols_used_for_filter = make_subscript_comparison_marks(
        row_slice, args.get("slice1_filter_labels", []), "column"
    )

    maybe_row = args.get("slice1_values")
    # TODO: indexers can be more types than just str and int e.g. datetimes
    if isinstance(maybe_row, (str, int)):
        row = util.positions_to_labels(
            maybe_row,
            df=before_df,
            slicer=step.slicer,
            axis="index",
        )

        return [
            *rows_used_for_filter,
            Map(from_=lhs("row", row), to=rhs_series()),
            # when slicing a row out of dataframe, the resulting series has
            # the df's column labels as the index. this means that the labels
            # are "transposed" so we don't draw arrows for this case.
        ]

    # df.iloc[:, 0]
    col = args.get("slice2_values")
    if isinstance(col, (str, int)):
        label = util.positions_to_labels(
            col,
            df=before_df,
            slicer=step.slicer,
            axis="columns",
        )
        return [
            *cols_used_for_filter,
            Map(from_=lhs("column", label), to=rhs_series()),
            *diff_rows(
                before_df,
                after_df,
                only_if_diff=(len(cols_used_for_filter) == 0),
            ),
        ]

    return []



"""
just for myself
"""
# type: ignore
# flake8: noqa

import prettyprinter  # type: ignore
from prettyprinter import cpprint  # type: ignore
from prettyprinter.prettyprinter import IMPLICIT_MODULES
import simplejson as json
from pandas_tutor.diagram import encode_dataclasses

from pandas_tutor.parse_nodes import Call, SortValuesCall  # type: ignore

from .parse import parse, test_logger, test_parser
from .run import run
from .__main__ import make_tutor_spec, make_tutor_spec_py

prettyprinter.install_extras(include=["dataclasses", "python", "numpy"])

# https://github.com/tommikaikkonen/prettyprinter/issues/27#issuecomment-451515061
IMPLICIT_MODULES.add("pandas_tutor.parse_nodes")
IMPLICIT_MODULES.add("pandas_tutor.util")

shorten_df = True

file_to_read = "parse_golden/assignment_last_line_01"


def p(obj):
    cpprint(obj, indent=2, ribbon_width=80)


if __name__ == "__main__":
    from pathlib import Path

    code = (Path(__file__).parent / f"tests/{file_to_read}.py").read_text()
    root = test_parser(code)
    # test_logger(code)
    #     print(code)
    #     print('\n--------------\n')

    # if shorten_df:
    #     for diagram in spec:
    #         lhs = diagram['data_frame']['lhs']
    #         rhs = diagram['data_frame']['rhs']
    #         lhs['data'] = len(lhs['data'])
    #         rhs['data'] = len(rhs['data'])

    p(root)
    # p(run(root))
    # spec = make_tutor_spec(code)
    # p(spec)

    print("\n---------------------------------------------------------\n")



# https://github.com/SamLau95/pandas_tutor/issues/86
# adapted from Thinking in Pandas
import pandas as pd

restaurants = pd.MultiIndex.from_tuples(
    (("Diner", (4, 2)), ("Pandas", (5, 4))),
    names=["restaurant", "location"],
)
inspections = pd.MultiIndex.from_tuples(
    ((0, "score"), (0, "date"), (1, "score"), (1, "date")),
    names=["inspection", None],
)

restaurant_inspections = pd.DataFrame(
    [[90, "02/18", 100, "05/18"], [55, "04/18", 76, "01/18"]],
    index=restaurants,
    columns=inspections,
)

restaurant_inspections.stack().stack().unstack().unstack().unstack().unstack()



import pandas as pd

df_list = [[1, 2, None], [1, None, 4], [2, 1, 4], [100, 200, -5], [1, 2, None]]
x = pd.DataFrame(df_list, columns=["a", "b", "c"])

# test groupby with NaN rows in keys
x.groupby(by=["c"], dropna=False).sum()



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from typing import Optional

import libcst as cst
from libcst.helpers.expression import get_full_name_for_node


def get_absolute_module_for_import(
    current_module: Optional[str], import_node: cst.ImportFrom
) -> Optional[str]:
    # First, let's try to grab the module name, regardless of relative status.
    module = import_node.module
    module_name = get_full_name_for_node(module) if module is not None else None
    # Now, get the relative import location if it exists.
    num_dots = len(import_node.relative)
    if num_dots == 0:
        # This is an absolute import, so the module is correct.
        return module_name
    if current_module is None:
        # We don't actually have the current module available, so we can't compute
        # the absolute module from relative.
        return None
    # We have the current module, as well as the relative, let's compute the base.
    modules = current_module.split(".")
    if len(modules) < num_dots:
        # This relative import goes past the base of the repository, so we can't calculate it.
        return None
    base_module = ".".join(modules[:-num_dots])
    # Finally, if the module name was supplied, append it to the end.
    if module_name is not None:
        # If we went all the way to the top, the base module should be empty, so we
        # should return the relative bit as absolute. Otherwise, combine the base
        # module and module name using a dot separator.
        base_module = (
            f"{base_module}.{module_name}" if len(base_module) > 0 else module_name
        )
    # If they tried to import all the way to the root, return None. Otherwise,
    # return the module itself.
    return base_module if len(base_module) > 0 else None


def get_absolute_module_for_import_or_raise(
    current_module: Optional[str], import_node: cst.ImportFrom
) -> str:
    module = get_absolute_module_for_import(current_module, import_node)
    if module is None:
        raise Exception(f"Unable to compute absolute module for {import_node}")
    return module



# https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html
import pandas as pd
df_list = [[1, 2, 3], [1, None, 4], [2, 1, 3], [1, 2, 2]]
df_dropna = pd.DataFrame(df_list, columns=["a", "b", "c"])
df_dropna.groupby(by=["b"])



"""Project metadata.

This information is used in setup.py as well as in doc/source/conf.py.

"""

project_name = 'Pympler'
version      = '0.9'
url          = 'https://github.com/pympler/pympler'
license      = 'Apache License, Version 2.0'
author       = 'Jean Brouwers, Ludwig Haehne, Robert Schuppenies'
author_email = 'pympler-dev@googlegroups.com'
copyright    = '2008-2020, ' + author
description  = ('A development tool to measure, monitor and analyze '
                'the memory behavior of Python objects.')
long_description = '''
Pympler is a development tool to measure, monitor and analyze the
memory behavior of Python objects in a running Python application.

By pympling a Python application, detailed insight in the size and
the lifetime of Python objects can be obtained.  Undesirable or
unexpected runtime behavior like memory bloat and other "pymples"
can easily be identified.

Pympler integrates three previously separate projects into a single,
comprehensive profiling tool. Asizeof provides basic size information
for one or several Python objects, muppy is used for on-line
monitoring of a Python application and the class tracker provides
off-line analysis of the lifetime of selected Python objects. A
web profiling frontend exposes process statistics, garbage
visualisation and class tracker statistics.

Pympler is written entirely in Python, with no dependencies to
external libraries. It has been tested with Python 2.7, 3.5, 3.6,
3.7, 3.8, 3.9 on Linux, Windows and MacOS X.
'''



import numpy as np
import pandas as pd

dates = pd.date_range("20130101", periods=6)

df = pd.DataFrame(np.random.randn(6, 4), index=dates, columns=list("ABCD"))

df2 = df.copy()

# should NOT be parsed as a subscript
df2["E"] = ["one", "one", "two", "three", "four", "three"]



# https://github.com/SamLau95/pandas_tutor/issues/86
# https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.unstack.html

import pandas as pd
import numpy as np

index = pd.MultiIndex.from_tuples([('one', 'a'), ('one', 'b'),
                                   ('two', 'a'), ('two', 'b')])
s = pd.Series(np.arange(1.0, 5.0), index=index)
                
# test cases (uncomment one at a time)
# s.unstack(level=-1)
#s.unstack(level=0)

# we don't draw marks for the second unstack since it goes from
# dataframe -> series 
s.unstack(level=0).unstack()


import pandas as pd
import numpy as np

# this uses about 3 MB
big = pd.DataFrame({'numbers': np.arange(400_000)})

# but we sample it to make it smaller
big = big.sample(n=100, random_state=42)

big.head()



"""
utils only used for making marks
"""
import itertools
from typing import Callable, Iterable, List, Tuple

import pandas as pd

from .. import util
from ..diagram import (
    Anchor,
    AxisPos,
    CellPos,
    Drop,
    IndexLevel,
    IndexLevelPos,
    Map,
    MapSet,
    Mark,
    PosPair,
    Selection,
    SeriesPos,
    TablePos,
    Using,
)
from ..parse_nodes import Axis
from ..util import Label, LabelPair


def diff_dfs(df1: pd.DataFrame, df2: pd.DataFrame):
    """
    when we just want to draw arrows between different rows and cols without
    special highlights. only outputs when there is at least one mismatching row
    / col
    """
    rows = diff_rows(df1, df2)
    cols = diff_cols(df1, df2)
    return [*cols, *rows]


def diff_rows(df1: util.HasIndex, df2: util.HasIndex, only_if_diff=True):
    """
    when we just want to draw arrows between different rows and cols without
    special highlights.
    """
    row_matches = util.match_rows(df1, df2, only_if_diff)
    return make_maps(row_matches, "row")


def diff_cols(df1: pd.DataFrame, df2: pd.DataFrame, only_if_diff=True):
    """
    when we just want to draw arrows between different rows and cols without
    special highlights.
    """
    col_matches = util.match_cols(df1, df2, only_if_diff)
    return make_maps(col_matches, "column")


def no_marks(*args) -> List[Mark]:
    # print(f'Unknown mark for {step.type_}')
    return []


def selection(axis: Axis, other=False) -> Selection:
    if other:
        return "column" if axis == "index" else "row"
    return "row" if axis == "index" else "column"


def make_usings(
    labels: Iterable, select: Selection, anchor: Anchor = "lhs"
) -> List[Mark]:
    """
    shorthand to make a highlight for each column/row in labels
    """
    return [Using(AxisPos(anchor, select, label)) for label in labels]


def make_maps(labels: Iterable, select: Selection) -> List[Mark]:
    """
    shorthand when index values don't change, which is most of the time
    """
    return [
        Map(from_=lhs(select, label), to=rhs(select, label)) for label in labels
    ]


def make_drops(
    labels: Iterable, select: Selection, anchor: Anchor = "lhs"
) -> List[Mark]:
    """
    shorthand for crossouts
    """
    return [Drop(AxisPos(anchor, select, label)) for label in labels]


def using_and_map(left_pos: TablePos, right_pos: TablePos) -> List[Mark]:
    """Map left to right and Using both"""
    return [Using(left_pos), Using(right_pos), Map(left_pos, right_pos)]


def lhs(select: Selection, label: Label) -> AxisPos:
    """shorthand for a column/row in lhs"""
    return AxisPos("lhs", select, label)


def rhs(select: Selection, label: Label) -> AxisPos:
    """shorthand for a column/row in rhs"""
    return AxisPos("rhs", select, label)


def lhs2(select: Selection, label: Label) -> AxisPos:
    """shorthand for a column/row in lhs2"""
    return AxisPos("lhs2", select, label)


def lhs_index(select: Selection, level: IndexLevel) -> IndexLevelPos:
    """shorthand for an index level in lhs"""
    return IndexLevelPos("lhs", select, level)


def rhs_index(select: Selection, level: IndexLevel) -> IndexLevelPos:
    """shorthand for an index level in rhs"""
    return IndexLevelPos("rhs", select, level)


def lhs2_index(select: Selection, level: IndexLevel) -> IndexLevelPos:
    """shorthand for an index level in lhs2"""
    return IndexLevelPos("lhs2", select, level)


def lhs_series() -> SeriesPos:
    """shorthand for the lhs series"""
    return SeriesPos("lhs")


def rhs_series() -> SeriesPos:
    """shorthand for the rhs series"""
    return SeriesPos("rhs")


def lhs2_series() -> SeriesPos:
    """shorthand for the lhs2 series"""
    return SeriesPos("lhs2")


def by_row(pair: Tuple[CellPos, CellPos]) -> Label:
    """grouper for CellPos pairs. returns the row label of the original cell"""
    cell, _ = pair
    return cell.row


def by_column(pair: Tuple[CellPos, CellPos]) -> Label:
    """grouper for CellPos pairs. returns the col label of the original cell"""
    cell, _ = pair
    return cell.column


def by_result_cell(pair: Tuple[CellPos, CellPos]) -> LabelPair:
    """
    grouper for CellPos pairs. returns the row, col pair for resulting cell
    """
    _, cell = pair
    return cell.row, cell.column


def make_map_sets(pairs: Iterable[PosPair], key: Callable) -> List[Mark]:
    """groups pairs by key, then makes a map set for each group"""
    pairs = sorted(pairs, key=key)
    return [
        MapSet([Map(from_, to) for from_, to in g])
        for _, g in itertools.groupby(pairs, key=key)
    ]


##############################################################################
# debugging
##############################################################################


def print_cell_sets(val: List[MapSet]) -> None:
    """helper to print mapsets of CellPos"""
    for i, mapset in enumerate(val):
        print(f"{i}:")
        for map_mark in mapset.maps:
            left = (
                f"{str(map_mark.from_.row)}, "  # type: ignore
                f"{str(map_mark.from_.column)}"  # type: ignore
            )
            right = (
                f"{str(map_mark.to.row)}, "  # type: ignore
                f"{str(map_mark.to.column)}"  # type: ignore
            )
            print(f"  {left} -> {right}")


def print_axis_sets(val: List[MapSet]) -> None:
    """helper to print mapsets of AxisPos"""
    for i, mapset in enumerate(val):
        print(f"{i}:")
        for map_mark in mapset.maps:
            left = (
                f"({str(map_mark.from_.anchor)}) "  # type: ignore
                f"{str(map_mark.from_.label)}"  # type: ignore
            )
            right = (
                f"({str(map_mark.to.anchor)}) "  # type: ignore
                f"{str(map_mark.to.label)}"  # type: ignore
            )
            print(f"  {left} -> {right}")



# https://pandas.pydata.org/pandas-docs/stable/user_guide/reshaping.html#reshaping-by-pivoting-dataframe-objects
import pandas as pd
import io

csv = """
foo,bar,baz,zoo
one,A,1,x
one,B,2,y
one,C,3,z
two,A,4,q
two,B,5,w
two,C,6,t
"""

df = pd.read_csv(io.StringIO(csv))

df.pivot(columns=["bar"], values=["baz", "foo"])



# https://pandas.pydata.org/docs/reference/api/pandas.melt.html
import pandas as pd

df = pd.DataFrame({'A': {0: 'a', 1: 'b', 2: 'c'},
                   'B': {0: 1, 1: 3, 2: 5},
                   'C': {0: 2, 1: 4, 2: 6}})

# sets variable and value column names
df.melt(id_vars='A', value_vars=['C', 'B'],
        var_name='hello', value_name='world')



"""
utilities
"""
from __future__ import annotations

import base64
import dataclasses
import io
import itertools
import warnings
from collections.abc import Iterable, Sequence
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Hashable,
    List,
    Tuple,
    TypeVar,
    Union,
    cast,
    overload,
)
from warnings import warn

import numpy as np
import pandas as pd
import pandas.core.groupby.base
from pandas.core.groupby.generic import DataFrameGroupBy, SeriesGroupBy
from pandas.core.groupby.groupby import GroupBy
from pandas.core.reshape.merge import _MergeOperation

# Literal only added in Python 3.8
from typing_extensions import Literal, TypeGuard

if TYPE_CHECKING:
    from pandas_tutor.diagram import MapSet  # noqa: F401

# functions that aggregate groups together
pd_agg_funcs: List[str] = pandas.core.groupby.base.reduction_kernels

T = TypeVar("T")

Axis = Literal["index", "columns"]
Slicer = Literal["loc", "iloc", None]

# A Label is a value that came from pandas Index. It's a tuple of values
# if we have a multi-index, or a single value otherwise.
Label = Hashable
LabelPair = Tuple[Label, Label]

HasIndex = Union[pd.DataFrame, pd.Series]

Groups = Dict[Union[str, tuple], pd.Index]

JSONScalar = Union[int, float, str, bool, None]

# placeholder column for pd.Series
SERIES = "pandas.Series"


def first(iterable: Iterable[T]) -> T:
    """the first element of an iterable."""
    return next(iter(iterable))


def second(iterable: Iterable[T]) -> T:
    """the second element of an iterable."""
    it = iter(iterable)
    next(it)
    return next(it)


def mapt(fn, *args):
    "map(fn, *args) and return the result as a tuple."
    return tuple(map(fn, *args))


def flatmap(fn, *args):
    "map(fn, *args) and return the result as a flattened iterable."
    return itertools.chain.from_iterable(map(fn, *args))


def is_list_like(obj: Any) -> bool:
    """
    checks whether obj is a list-like. we need this because we don't usually
    want to do list(string), but we want to convert other types of list-like
    things to lists
    """
    return not isinstance(obj, str) and isinstance(obj, Iterable)


def listify(obj: Any) -> List:
    """if obj is a scalar, returns [obj]"""
    return obj if is_list_like(obj) else [obj]


def tuplify(obj: Any) -> Tuple:
    """if obj is a scalar, returns (obj, )"""
    return obj if is_list_like(obj) else (obj,)


@overload
def unwrap(obj: Tuple[T]) -> Union[Tuple[T], T]:
    ...


@overload
def unwrap(obj: List[T]) -> Union[List[T], T]:
    ...


@overload
def unwrap(obj: Label) -> Label:
    ...


def unwrap(obj):
    """unwrap obj if single element"""
    return obj if not is_list_like(obj) else obj[0] if len(obj) == 1 else obj


def split_by(
    seq: Sequence[T], indexes: Sequence[int]
) -> Tuple[Tuple[T, ...], Tuple[T, ...]]:
    """
    returns two tuples, one with the elements at the given indexes, and one
    with the rest of the elements.
    """
    # can't use sets since we need to preserve order
    not_in = [i for i in range(len(seq)) if i not in indexes]
    return tuple(seq[i] for i in indexes), tuple(seq[i] for i in not_in)


@dataclasses.dataclass
class CodePosition:
    """
    points to a location within the original code string. both lines and
    columns are 0-indexed
    """

    line: int
    ch: int

    def __mod__(self, other: CodePosition):
        """self % other is the position relative to other.line"""
        return CodePosition(self.line - other.line, self.ch)

    def __lt__(self, other: CodePosition):
        return (
            self.line < other.line
            if self.line != other.line
            else self.ch < other.ch
        )

    def __gt__(self, other: CodePosition):
        return (
            self.line > other.line
            if self.line != other.line
            else self.ch > other.ch
        )

    def __str__(self):
        return f"({self.line}:{self.ch})"


@dataclasses.dataclass
class CodeRange:
    """
    points to a code range within the original code string. both lines and
    columns are 0-indexed.

    these are used to highlight code fragments in the frontend
    """

    start: CodePosition
    end: CodePosition

    def __sub__(self, other: CodeRange):
        """
        a range within this CodeRange that doesn't overlap with other. similar
        to set difference. assumes the CodeRanges only partially overlap
        """
        # non-overlapping
        if self.end < other.start or other.end < self.start:
            return self

        # cut off left tail, common case
        if self.end > other.end:
            return CodeRange(other.end, self.end)

        # cut off right tail
        if self.start < other.start:
            return CodeRange(self.start, other.start)

        return self

    def __or__(self, other: CodeRange):
        """
        minimum CodeRange that contains both self and other
        """
        return CodeRange(
            start=self.start if self.start < other.start else other.start,
            end=self.end if self.end > other.end else other.end,
        )

    def __mod__(self, pos: CodePosition):
        """self % pos is the range relative to the starting line of pos"""
        return CodeRange(self.start % pos, self.end % pos)

    def __str__(self):
        return f"{self.start} -> {self.end}"


##############################################################################
# pandas
##############################################################################


@overload
def positions_to_labels(
    positions: Union[int, Label],
    df: HasIndex,
    slicer: Slicer = "iloc",
    axis: Axis = "index",
) -> Label:
    ...


@overload
def positions_to_labels(  # noqa: F811
    positions: list,  # type: ignore
    df: HasIndex,
    slicer: Slicer = "iloc",
    axis: Axis = "index",
) -> List[Label]:
    ...


def positions_to_labels(  # noqa: F811
    positions,
    df,
    slicer: Slicer = "iloc",
    axis: Axis = "index",
):
    """
    convert positional indexes like [2, 3, 0] to labels.
    doesn't do anything if slicer isn't iloc.
    if positions is a single number, also returns a single label.
    """
    if slicer != "iloc":
        return positions
    if axis != "index" and isinstance(df, pd.Series):
        warn("tried to convert column labels for a series")
        return positions

    labels = cast(pd.Index, df.columns if axis == "columns" else df.index)
    return labels[positions]


def match_rows(df1: HasIndex, df2: HasIndex, only_if_diff=True) -> pd.Index:
    """
    find all matching row labels between df1 and df2. if only_if_diff=True
    (default), then return empty index when df1 has same rows as df2.
    """
    # TODO: doesn't handle duplicate values in an index properly, since:
    # >>> a = pd.Index([2, 2])
    # >>> a.intersection(a)
    # Index([2])
    matches = df1.index.intersection(df2.index)  # type: ignore
    return (
        pd.Index([], dtype="int64")
        if (len(matches) == len(df1.index) and only_if_diff)
        else matches
    )


def match_cols(
    df1: pd.DataFrame, df2: pd.DataFrame, only_if_diff=True
) -> pd.Index:
    """
    find all matching col labels between df1 and df2. if only_if_diff=True
    (default), then return empty index when df1 has same cols as df2.
    """
    matches = df1.columns.intersection(df2.columns)
    return (
        pd.Index([], dtype="int64")
        if (len(matches) == len(df1.columns) and only_if_diff)
        else matches
    )


def is_series(obj: Any) -> TypeGuard[pd.Series]:
    return isinstance(obj, pd.Series)


def is_dataframe(obj: Any) -> TypeGuard[pd.DataFrame]:
    return isinstance(obj, pd.DataFrame)


def is_pd(obj: Any) -> TypeGuard[Union[pd.DataFrame, pd.Series]]:
    return is_series(obj) or is_dataframe(obj)


def is_multi(index: pd.Index) -> TypeGuard[pd.MultiIndex]:
    return isinstance(index, pd.MultiIndex)


def level_number(index: pd.Index, level: str | int) -> int:
    "converts a name of a level to the level's integer index if needed"
    return (
        index.names.index(level)
        if isinstance(level, str)
        else level % len(index.names)
    )


def first_iloc(index: pd.Index, label: Label) -> int:
    """
    returns the first position of a label in an index.
    """
    for i, lab in enumerate(index):
        if lab == label:
            return i
    return -1


##############################################################################
# multiindex
##############################################################################

# pair of multi-index labels
_MultiLabelPair = Tuple[Tuple[Label, ...], Tuple[Label, ...]]


def cell_labels(
    index1: Iterable[Label], index2: Iterable[Label]
) -> Tuple[LabelPair]:
    """returns (row, column) labels"""
    return tuple(itertools.product(index1, index2))  # type: ignore


def push_levels(
    from_: Iterable[Label], to: Iterable[Label], levels: List[int]
) -> Iterable[Tuple[LabelPair, LabelPair]]:
    """
    pushes levels from_ -> to. used to map cells during reshaping operations.
    returns original and new label pairs.
    """
    orig = cell_labels(from_, to)

    # if we're pushing into a series, we need to handle the placeholder
    iter_to = to if SERIES not in to else [tuple()]

    # wrap values in tuples to make iteration easier
    wrapped_orig = cast(
        List[_MultiLabelPair],
        list(itertools.product(map(tuplify, from_), map(tuplify, iter_to))),
    )

    new: List[LabelPair] = []
    for left, right in wrapped_orig:
        move, stay = split_by(left, levels)
        new.append(mapt(unwrap, (stay, (*right, *move))))

    return zip(orig, new)


def push_level(from_: Label, to: Label, levels: List[int]) -> LabelPair:
    """pushes levels from_ -> to for one label pair"""
    # if we're pushing into a series, we need to handle the placeholder
    to = to if to != SERIES else tuple()
    left = tuplify(from_)
    right = tuplify(to)
    move, stay = split_by(left, levels)
    return mapt(unwrap, (stay, (*right, *move)))


##############################################################################
# groupby
##############################################################################


@overload
def ungroup(obj: Union[SeriesGroupBy, pd.Series]) -> pd.Series:
    ...


@overload
def ungroup(  # type: ignore # noqa: F811
    obj: Union[DataFrameGroupBy, pd.DataFrame]  # noqa: F811
) -> pd.DataFrame:
    ...


def ungroup(obj):  # noqa: F811
    """
    undos a groupby back into original val. if obj isn't grouped, returns obj
    """
    if isinstance(obj, (SeriesGroupBy, DataFrameGroupBy)):
        # uses a private attribute...hopefully won't break later :)
        return obj._selected_obj
    return obj

    # slower fallback
    # return groupby.transform(lambda x: x)


def grouping_labels(groupby: GroupBy) -> List[Label]:
    """gets ['hello', 'world'] from df.groupby(['hello', 'world'])"""
    # NOTE: when grouping by unnamed sequences, names will contain None
    # >>> full.groupby([test, test2]).grouper.names
    # [None, None]
    return groupby.grouper.names


def get_groups(groupby: Union[SeriesGroupBy, DataFrameGroupBy]) -> Groups:
    """
    gets mapping of group keys -> dataframe labels.
    """
    # when the group keys includes NaN, groupby.groups freaks out, so we use a
    # workaround by getting the group indices first, then recovering the labels
    try:
        return cast(Groups, groupby.groups)
    except ValueError:
        index = ungroup(groupby).index
        groups = {
            key: index[indices] for key, indices in groupby.indices.items()
        }
        return cast(Groups, groups)


##############################################################################
# merge
##############################################################################


def get_join_info(*args, **kwargs) -> Tuple[pd.Index, pd.Index, pd.Index]:
    """
    call using same signature as pd.merge. returns (index of merged df, indexer
    for lhs, indexer for lhs2). uses pandas internals.

    >>> op = _MergeOperation(baby, nyt2,
    ...                      left_on='Name', right_on='nyt_name', how='left')
    >>> join_index, left_indexer, right_indexer = op._get_join_info()
    >>> join_index, left_indexer, right_indexer
    (Int64Index([0, 1, 2, 3, 4], dtype='int64'),
     array([1, 2, 3, 0, 4]),
     array([ 1,  0,  0, -1, -1]))
    """
    op = _MergeOperation(*args, **kwargs)
    return op._get_join_info()  # type: ignore


##############################################################################
# serializing
##############################################################################


def is_plottable(obj: Any) -> bool:
    fig = obj.figure if hasattr(obj, "figure") else obj
    return hasattr(fig, "savefig")


def base64_encode_plot(fig_or_axes: Any) -> str:
    """
    saves plot as a gzipped, base64 encoded png
    """
    if not is_plottable(fig_or_axes):
        return ""

    fig = fig_or_axes.figure if hasattr(fig_or_axes, "figure") else fig_or_axes

    # saves figure as base64 encoded string
    with io.BytesIO() as buf:
        fig.savefig(buf, format="png", bbox_inches="tight")
        buf.seek(0)
        # set mtime=0 to get deterministic gzips for testing
        # zipped = gzip.compress(buf.read(), mtime=0)
        return base64.b64encode(buf.read()).decode()


def json_scalar(obj: Any) -> JSONScalar:
    """
    transforms special pandas / numpy value to a value that can be
    serialized to json
    """
    if pd.isnull(obj):
        return None
    if isinstance(obj, np.integer):
        return int(obj)
    if isinstance(obj, np.floating):
        return float(obj)
    if isinstance(obj, np.ndarray):
        return obj.tolist()
    if isinstance(obj, pd.Timestamp) or isinstance(obj, pd.Timedelta):
        return str(obj)

    # groupby().plot() creates a Series with AxesSubplots as data values.
    # weird!
    if obj.__class__.__name__ == "AxesSubplot":
        return str(obj)

    return obj


def index_data(index: pd.Index) -> List[JSONScalar]:
    return [json_scalar(val) for val in index.to_list()]


def series_data(series: pd.Series) -> List[JSONScalar]:
    return [json_scalar(val) for val in series.to_numpy()]


def df_data(df: pd.DataFrame) -> List[List[JSONScalar]]:
    return [[json_scalar(val) for val in row] for row in df.to_numpy()]


##############################################################################
# memory
##############################################################################


def mem_used(obj: Any) -> float:
    if isinstance(obj, pd.DataFrame):
        return obj.memory_usage(deep=True).sum()
    elif isinstance(obj, pd.Series):
        return obj.memory_usage(deep=True)
    else:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            from pympler.asizeof import asizeof
        return asizeof(obj)


KB = 2**10
MB = 2**20
MEM_LIMIT = 1 * MB


def mem_as_str(mem: float) -> str:
    if mem >= MB:
        # 2 decimal places
        return f"{mem / MB:.2f} MB"
    elif mem >= KB:
        return f"{mem / KB:.2f} KB"
    else:
        return f"{mem} B"


def too_much_mem_msg(mem: float):
    return (
        f"Your total data uses {mem_as_str(mem)} of memory, which exceeds "
        f"the maximum of {mem_as_str(MEM_LIMIT)} that this tool supports."
    )



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from typing import TYPE_CHECKING, Union

from libcst._flatten_sentinel import FlattenSentinel
from libcst._metadata_dependent import MetadataDependent
from libcst._removal_sentinel import RemovalSentinel
from libcst._typed_visitor import CSTTypedTransformerFunctions, CSTTypedVisitorFunctions
from libcst._types import CSTNodeT

if TYPE_CHECKING:
    # Circular dependency for typing reasons only
    from libcst._nodes.base import CSTNode  # noqa: F401


CSTVisitorT = Union["CSTTransformer", "CSTVisitor"]


class CSTTransformer(CSTTypedTransformerFunctions, MetadataDependent):
    """
    The low-level base visitor class for traversing a CST and creating an
    updated copy of the original CST. This should be used in conjunction with
    the :func:`~libcst.CSTNode.visit` method on a :class:`~libcst.CSTNode` to
    visit each element in a tree starting with that node, and possibly returning
    a new node in its place.

    When visiting nodes using a :class:`CSTTransformer`, the return value of
    :func:`~libcst.CSTNode.visit` will be a new tree with any changes made in
    :func:`~libcst.CSTTransformer.on_leave` calls reflected in its children.
    """

    def on_visit(self, node: "CSTNode") -> bool:
        """
        Called every time a node is visited, before we've visited its children.

        Returns ``True`` if children should be visited, and returns ``False``
        otherwise.
        """
        visit_func = getattr(self, f"visit_{type(node).__name__}", None)
        if visit_func is not None:
            retval = visit_func(node)
        else:
            retval = True
        # Don't visit children IFF the visit function returned False.
        return False if retval is False else True

    def on_leave(
        self, original_node: CSTNodeT, updated_node: CSTNodeT
    ) -> Union[CSTNodeT, RemovalSentinel, FlattenSentinel[CSTNodeT]]:
        """
        Called every time we leave a node, after we've visited its children. If
        the :func:`~libcst.CSTTransformer.on_visit` function for this node returns
        ``False``, this function will still be called on that node.

        ``original_node`` is guaranteed to be the same node as is passed to
        :func:`~libcst.CSTTransformer.on_visit`, so it is safe to do state-based
        checks using the ``is`` operator. Modifications should always be performed
        on the ``updated_node`` so as to not overwrite changes made by child
        visits.

        Returning :attr:`RemovalSentinel.REMOVE` indicates that the node should be
        removed from its parent. This is not always possible, and may raise an
        exception if this node is required. As a convenience, you can use
        :func:`RemoveFromParent` as an alias to :attr:`RemovalSentinel.REMOVE`.
        """
        leave_func = getattr(self, f"leave_{type(original_node).__name__}", None)
        if leave_func is not None:
            updated_node = leave_func(original_node, updated_node)

        return updated_node

    def on_visit_attribute(self, node: "CSTNode", attribute: str) -> None:
        """
        Called before a node's child attribute is visited and after we have called
        :func:`~libcst.CSTTransformer.on_visit` on the node. A node's child
        attributes are visited in the order that they appear in source that this
        node originates from.
        """
        visit_func = getattr(self, f"visit_{type(node).__name__}_{attribute}", None)
        if visit_func is not None:
            visit_func(node)

    def on_leave_attribute(self, original_node: "CSTNode", attribute: str) -> None:
        """
        Called after a node's child attribute is visited and before we have called
        :func:`~libcst.CSTTransformer.on_leave` on the node.

        Unlike :func:`~libcst.CSTTransformer.on_leave`, this function does
        not allow modifications to the tree and is provided solely for state
        management.
        """
        leave_func = getattr(
            self, f"leave_{type(original_node).__name__}_{attribute}", None
        )
        if leave_func is not None:
            leave_func(original_node)


class CSTVisitor(CSTTypedVisitorFunctions, MetadataDependent):
    """
    The low-level base visitor class for traversing a CST. This should be used in
    conjunction with the :func:`~libcst.CSTNode.visit` method on a
    :class:`~libcst.CSTNode` to visit each element in a tree starting with that
    node. Unlike :class:`CSTTransformer`, instances of this class cannot modify
    the tree.

    When visiting nodes using a :class:`CSTVisitor`, the return value of
    :func:`~libcst.CSTNode.visit` will equal the passed in tree.
    """

    def on_visit(self, node: "CSTNode") -> bool:
        """
        Called every time a node is visited, before we've visited its children.

        Returns ``True`` if children should be visited, and returns ``False``
        otherwise.
        """
        visit_func = getattr(self, f"visit_{type(node).__name__}", None)
        if visit_func is not None:
            retval = visit_func(node)
        else:
            retval = True
        # Don't visit children IFF the visit function returned False.
        return False if retval is False else True

    def on_leave(self, original_node: "CSTNode") -> None:
        """
        Called every time we leave a node, after we've visited its children. If
        the :func:`~libcst.CSTVisitor.on_visit` function for this node returns
        ``False``, this function will still be called on that node.
        """
        leave_func = getattr(self, f"leave_{type(original_node).__name__}", None)
        if leave_func is not None:
            leave_func(original_node)

    def on_visit_attribute(self, node: "CSTNode", attribute: str) -> None:
        """
        Called before a node's child attribute is visited and after we have called
        :func:`~libcst.CSTTransformer.on_visit` on the node. A node's child
        attributes are visited in the order that they appear in source that this
        node originates from.
        """
        visit_func = getattr(self, f"visit_{type(node).__name__}_{attribute}", None)
        if visit_func is not None:
            visit_func(node)

    def on_leave_attribute(self, original_node: "CSTNode", attribute: str) -> None:
        """
        Called after a node's child attribute is visited and before we have called
        :func:`~libcst.CSTVisitor.on_leave` on the node.
        """
        leave_func = getattr(
            self, f"leave_{type(original_node).__name__}_{attribute}", None
        )
        if leave_func is not None:
            leave_func(original_node)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

"""
Parso doesn't attempt to parse (or even emit tokens for) whitespace or comments that
isn't syntatically important. Instead, we're just given the whitespace as a "prefix" of
the token.

However, in our CST, whitespace is gathered into far more detailed objects than a simple
str.

Fortunately this isn't hard for us to parse ourselves, so we just use our own
hand-rolled recursive descent parser.
"""

from typing import List, Optional, Sequence, Tuple, Union

from libcst._nodes.whitespace import (
    COMMENT_RE,
    NEWLINE_RE,
    SIMPLE_WHITESPACE_RE,
    Comment,
    EmptyLine,
    Newline,
    ParenthesizedWhitespace,
    SimpleWhitespace,
    TrailingWhitespace,
)
from libcst._parser.types.config import BaseWhitespaceParserConfig
from libcst._parser.types.whitespace_state import WhitespaceState as State

# BEGIN PARSER ENTRYPOINTS


def parse_simple_whitespace(
    config: BaseWhitespaceParserConfig, state: State
) -> SimpleWhitespace:
    # The match never fails because the pattern can match an empty string
    lines = config.lines
    # pyre-fixme[16]: Optional type has no attribute `group`.
    ws_line = SIMPLE_WHITESPACE_RE.match(lines[state.line - 1], state.column).group(0)
    ws_line_list = [ws_line]
    while "\\" in ws_line:
        # continuation character
        state.line += 1
        state.column = 0
        ws_line = SIMPLE_WHITESPACE_RE.match(lines[state.line - 1], state.column).group(
            0
        )
        ws_line_list.append(ws_line)

    # TODO: we could special-case the common case where there's no continuation
    # character to avoid list construction and joining.

    # once we've finished collecting continuation characters
    state.column += len(ws_line)
    return SimpleWhitespace("".join(ws_line_list))


def parse_empty_lines(
    config: BaseWhitespaceParserConfig,
    state: State,
    *,
    override_absolute_indent: Optional[str] = None,
) -> Sequence[EmptyLine]:
    # If override_absolute_indent is true, then we need to parse all lines up
    # to and including the last line that is indented at our level. These all
    # belong to the footer and not to the next line's leading_lines. All lines
    # that have indent=False and come after the last line where indent=True
    # do not belong to this node.
    state_for_line = State(
        state.line, state.column, state.absolute_indent, state.is_parenthesized
    )
    lines: List[Tuple[State, EmptyLine]] = []
    while True:
        el = _parse_empty_line(
            config, state_for_line, override_absolute_indent=override_absolute_indent
        )
        if el is None:
            break

        # Store the updated state with the element we parsed. Then make a new state
        # clone for the next element.
        lines.append((state_for_line, el))
        state_for_line = State(
            state_for_line.line,
            state_for_line.column,
            state.absolute_indent,
            state.is_parenthesized,
        )

    if override_absolute_indent is not None:
        # We need to find the last element that is indented, and then split the list
        # at that point.
        for i in range(len(lines) - 1, -1, -1):
            if lines[i][1].indent:
                lines = lines[: (i + 1)]
                break
        else:
            # We didn't find any lines, throw them all away
            lines = []

    if lines:
        # Update the state line and column to match the last line actually parsed.
        final_state: State = lines[-1][0]
        state.line = final_state.line
        state.column = final_state.column
    return [r[1] for r in lines]


def parse_trailing_whitespace(
    config: BaseWhitespaceParserConfig, state: State
) -> TrailingWhitespace:
    trailing_whitespace = _parse_trailing_whitespace(config, state)
    if trailing_whitespace is None:
        raise Exception(
            "Internal Error: Failed to parse TrailingWhitespace. This should never "
            + "happen because a TrailingWhitespace is never optional in the grammar, "
            + "so this error should've been caught by parso first."
        )
    return trailing_whitespace


def parse_parenthesizable_whitespace(
    config: BaseWhitespaceParserConfig, state: State
) -> Union[SimpleWhitespace, ParenthesizedWhitespace]:
    if state.is_parenthesized:
        # First, try parenthesized (don't need speculation because it either
        # parses or doesn't modify state).
        parenthesized_whitespace = _parse_parenthesized_whitespace(config, state)
        if parenthesized_whitespace is not None:
            return parenthesized_whitespace
    # Now, just parse and return a simple whitespace
    return parse_simple_whitespace(config, state)


# END PARSER ENTRYPOINTS
# BEGIN PARSER INTERNAL PRODUCTIONS


def _parse_empty_line(
    config: BaseWhitespaceParserConfig,
    state: State,
    *,
    override_absolute_indent: Optional[str] = None,
) -> Optional[EmptyLine]:
    # begin speculative parsing
    speculative_state = State(
        state.line, state.column, state.absolute_indent, state.is_parenthesized
    )
    try:
        indent = _parse_indent(
            config, speculative_state, override_absolute_indent=override_absolute_indent
        )
    except Exception:
        # We aren't on a new line, speculative parsing failed
        return None
    whitespace = parse_simple_whitespace(config, speculative_state)
    comment = _parse_comment(config, speculative_state)
    newline = _parse_newline(config, speculative_state)
    if newline is None:
        # speculative parsing failed
        return None
    # speculative parsing succeeded
    state.line = speculative_state.line
    state.column = speculative_state.column
    # don't need to copy absolute_indent/is_parenthesized because they don't change.
    return EmptyLine(indent, whitespace, comment, newline)


def _parse_indent(
    config: BaseWhitespaceParserConfig,
    state: State,
    *,
    override_absolute_indent: Optional[str] = None,
) -> bool:
    """
    Returns True if indentation was found, otherwise False.
    """
    absolute_indent = (
        override_absolute_indent
        if override_absolute_indent is not None
        else state.absolute_indent
    )
    line_str = config.lines[state.line - 1]
    if state.column != 0:
        if state.column == len(line_str) and state.line == len(config.lines):
            # We're at EOF, treat this as a failed speculative parse
            return False
        raise Exception("Internal Error: Column should be 0 when parsing an indent.")
    if line_str.startswith(absolute_indent, state.column):
        state.column += len(absolute_indent)
        return True
    return False


def _parse_comment(
    config: BaseWhitespaceParserConfig, state: State
) -> Optional[Comment]:
    comment_match = COMMENT_RE.match(config.lines[state.line - 1], state.column)
    if comment_match is None:
        return None
    comment = comment_match.group(0)
    state.column += len(comment)
    return Comment(comment)


def _parse_newline(
    config: BaseWhitespaceParserConfig, state: State
) -> Optional[Newline]:
    # begin speculative parsing
    line_str = config.lines[state.line - 1]
    newline_match = NEWLINE_RE.match(line_str, state.column)
    if newline_match is not None:
        # speculative parsing succeeded
        newline_str = newline_match.group(0)
        state.column += len(newline_str)
        if state.column != len(line_str):
            raise Exception("Internal Error: Found a newline, but it wasn't the EOL.")
        if state.line < len(config.lines):
            # this newline was the end of a line, and there's another line,
            # therefore we should move to the next line
            state.line += 1
            state.column = 0
        if newline_str == config.default_newline:
            # Just inherit it from the Module instead of explicitly setting it.
            return Newline()
        else:
            return Newline(newline_str)
    else:  # no newline was found, speculative parsing failed
        return None


def _parse_trailing_whitespace(
    config: BaseWhitespaceParserConfig, state: State
) -> Optional[TrailingWhitespace]:
    # Begin speculative parsing
    speculative_state = State(
        state.line, state.column, state.absolute_indent, state.is_parenthesized
    )
    whitespace = parse_simple_whitespace(config, speculative_state)
    comment = _parse_comment(config, speculative_state)
    newline = _parse_newline(config, speculative_state)
    if newline is None:
        # Speculative parsing failed
        return None
    # Speculative parsing succeeded
    state.line = speculative_state.line
    state.column = speculative_state.column
    # don't need to copy absolute_indent/is_parenthesized because they don't change.
    return TrailingWhitespace(whitespace, comment, newline)


def _parse_parenthesized_whitespace(
    config: BaseWhitespaceParserConfig, state: State
) -> Optional[ParenthesizedWhitespace]:
    first_line = _parse_trailing_whitespace(config, state)
    if first_line is None:
        # Speculative parsing failed
        return None
    empty_lines = ()
    while True:
        empty_line = _parse_empty_line(config, state)
        if empty_line is None:
            # This isn't an empty line, so parse it below
            break
        empty_lines = empty_lines + (empty_line,)
    indent = _parse_indent(config, state)
    last_line = parse_simple_whitespace(config, state)
    return ParenthesizedWhitespace(first_line, empty_lines, indent, last_line)



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#
from libcst.codemod._cli import (
    ParallelTransformResult,
    diff_code,
    exec_transform_with_prettyprint,
    gather_files,
    parallel_exec_transform_with_prettyprint,
)
from libcst.codemod._codemod import Codemod
from libcst.codemod._command import (
    CodemodCommand,
    MagicArgsCodemodCommand,
    VisitorBasedCodemodCommand,
)
from libcst.codemod._context import CodemodContext
from libcst.codemod._runner import (
    SkipFile,
    SkipReason,
    TransformExit,
    TransformFailure,
    TransformResult,
    TransformSkip,
    TransformSuccess,
    transform_module,
)
from libcst.codemod._testing import CodemodTest
from libcst.codemod._visitor import ContextAwareTransformer, ContextAwareVisitor

__all__ = [
    "Codemod",
    "CodemodContext",
    "CodemodCommand",
    "VisitorBasedCodemodCommand",
    "MagicArgsCodemodCommand",
    "ContextAwareTransformer",
    "ContextAwareVisitor",
    "ParallelTransformResult",
    "TransformSuccess",
    "TransformFailure",
    "TransformExit",
    "SkipReason",
    "TransformSkip",
    "SkipFile",
    "TransformResult",
    "CodemodTest",
    "transform_module",
    "gather_files",
    "exec_transform_with_prettyprint",
    "parallel_exec_transform_with_prettyprint",
    "diff_code",
]



# Copyright (c) Facebook, Inc. and its affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.
#

"""
Provides everything needed to run a CodemodCommand.
"""

import traceback
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Sequence, Union

from libcst import PartialParserConfig, parse_module
from libcst.codemod._codemod import Codemod

# All datastructures defined in this class are pickleable so that they can be used
# as a return value with the multiprocessing module.


@dataclass(frozen=True)
class TransformSuccess:
    """
    A :class:`~libcst.codemod.TransformResult` used when the codemod was successful.
    Stores all the information we might need to display to the user upon success, as
    well as the transformed file contents.
    """

    #: All warning messages that were generated during the codemod.
    warning_messages: Sequence[str]

    #: The updated code, post-codemod.
    code: str


@dataclass(frozen=True)
class TransformFailure:
    """
    A :class:`~libcst.codemod.TransformResult` used when the codemod failed.
    Stores all the information we might need to display to the user upon a failure.
    """

    #: All warning messages that were generated before the codemod crashed.
    warning_messages: Sequence[str]

    #: The exception that was raised during the codemod.
    error: Exception

    #: The traceback string that was recorded at the time of exception.
    traceback_str: str


@dataclass(frozen=True)
class TransformExit:
    """
    A :class:`~libcst.codemod.TransformResult` used when the codemod was interrupted
    by the user (e.g. KeyboardInterrupt).
    """

    #: An empty list of warnings, included so that all
    #: :class:`~libcst.codemod.TransformResult` have a ``warning_messages`` attribute.
    warning_messages: Sequence[str] = ()


class SkipReason(Enum):
    """
    An enumeration of all valid reasons for a codemod to skip.
    """

    #: The module was skipped because we detected that it was generated code, and
    #: we were configured to skip generated files.
    GENERATED = "generated"

    #: The module was skipped because we detected that it was blacklisted, and we
    #: were configured to skip blacklisted files.
    BLACKLISTED = "blacklisted"

    #: The module was skipped because the codemod requested us to skip using the
    #: :class:`~libcst.codemod.SkipFile` exception.
    OTHER = "other"


@dataclass(frozen=True)
class TransformSkip:
    """
    A :class:`~libcst.codemod.TransformResult` used when the codemod requested to
    be skipped. This could be because it's a generated file, or due to filename
    blacklist, or because the transform raised :class:`~libcst.codemod.SkipFile`.
    """

    #: The reason that we skipped codemodding this module.
    skip_reason: SkipReason

    #: The description populated from the :class:`~libcst.codemod.SkipFile` exception.
    skip_description: str

    #: All warning messages that were generated before the codemod decided to skip.
    warning_messages: Sequence[str] = ()


class SkipFile(Exception):
    """
    Raise this exception to skip codemodding the current file.

    The exception message should be the reason for skipping.
    """


TransformResult = Union[
    TransformSuccess, TransformFailure, TransformExit, TransformSkip
]


def transform_module(
    transformer: Codemod, code: str, *, python_version: Optional[str] = None
) -> TransformResult:
    """
    Given a module as represented by a string and a :class:`~libcst.codemod.Codemod`
    that we wish to run, execute the codemod on the code and return a
    :class:`~libcst.codemod.TransformResult`. This should never raise an exception.
    On success, this returns a :class:`~libcst.codemod.TransformSuccess` containing
    any generated warnings as well as the transformed code. If the codemod is
    interrupted with a Ctrl+C, this returns a :class:`~libcst.codemod.TransformExit`.
    If the codemod elected to skip by throwing a :class:`~libcst.codemod.SkipFile`
    exception, this will return a :class:`~libcst.codemod.TransformSkip` containing
    the reason for skipping as well as any warnings that were generated before
    the codemod decided to skip. If the codemod throws an unexpected exception,
    this will return a :class:`~libcst.codemod.TransformFailure` containing the
    exception that occured as well as any warnings that were generated before the
    codemod crashed.
    """
    try:
        input_tree = parse_module(
            code,
            config=(
                PartialParserConfig(python_version=python_version)
                if python_version is not None
                else PartialParserConfig()
            ),
        )
        output_tree = transformer.transform_module(input_tree)
        return TransformSuccess(
            code=output_tree.code, warning_messages=transformer.context.warnings
        )
    except KeyboardInterrupt:
        return TransformExit()
    except SkipFile as ex:
        return TransformSkip(
            skip_description=str(ex),
            skip_reason=SkipReason.OTHER,
            warning_messages=transformer.context.warnings,
        )
    except Exception as ex:
        return TransformFailure(
            error=ex,
            traceback_str=traceback.format_exc(),
            warning_messages=transformer.context.warnings,
        )



# https://www.shanelynn.ie/pandas-iloc-loc-select-rows-and-columns-dataframe/
import pandas as pd
import io

csv = """
first_name,last_name,company_name,address,city,county,postal,phone1,phone2,email,web
Aleshia,Tomkiewicz,Alan D Rosenburg Cpa Pc,14 Taylor St,St. Stephens Ward,Kent,CT2 7PP,01835-703597,01944-369967,atomkiewicz@hotmail.com,http://www.alandrosenburgcpapc.co.uk
Evan,Zigomalas,Cap Gemini America,5 Binney St,Abbey Ward,Buckinghamshire,HP11 2AX,01937-864715,01714-737668,evan.zigomalas@gmail.com,http://www.capgeminiamerica.co.uk
France,Andrade,"Elliott, John W Esq",8 Moor Place,East Southbourne and Tuckton W,Bournemouth,BH6 3BE,01347-368222,01935-821636,france.andrade@hotmail.com,http://www.elliottjohnwesq.co.uk
Ulysses,Mcwalters,"Mcmahan, Ben L",505 Exeter Rd,Hawerby cum Beesby,Lincolnshire,DN36 5RP,01912-771311,01302-601380,ulysses@hotmail.com,http://www.mcmahanbenl.co.uk
Tyisha,Veness,Champagne Room,5396 Forth Street,Greets Green and Lyng Ward,West Midlands,B70 9DT,01547-429341,01290-367248,tyisha.veness@hotmail.com,http://www.champagneroom.co.uk
Eric,Rampy,"Thompson, Michael C Esq",9472 Lind St,Desborough,Northamptonshire,NN14 2GH,01969-886290,01545-817375,erampy@rampy.co.uk,http://www.thompsonmichaelcesq.co.uk
Marg,Grasmick,Wrangle Hill Auto Auct & Slvg,7457 Cowl St #70,Bargate Ward,Southampton,SO14 3TY,01865-582516,01362-620532,marg@hotmail.com,http://www.wranglehillautoauctslvg.co.uk
Laquita,Hisaw,In Communications Inc,20 Gloucester Pl #96,Chirton Ward,Tyne & Wear,NE29 7AD,01746-394243,01590-982428,laquita@yahoo.com,http://www.incommunicationsinc.co.uk
Lura,Manzella,Bizerba Usa Inc,929 Augustine St,Staple Hill Ward,South Gloucestershire,BS16 4LL,01907-538509,01340-713951,lura@hotmail.com,http://www.bizerbausainc.co.uk
Yuette,Klapec,Max Video,45 Bradfield St #166,Parwich,Derbyshire,DE6 1QN,01903-649460,01933-512513,yuette.klapec@klapec.co.uk,http://www.maxvideo.co.uk
Fernanda,Writer,K & R Associates Inc,620 Northampton St,Wilmington,Kent,DA2 7PP,01630-202053,01687-879391,fernanda@writer.co.uk,http://www.krassociatesinc.co.uk
Charlesetta,Erm,"Cain, John M Esq",5 Hygeia St,Loundsley Green Ward,Derbyshire,S40 4LY,01276-816806,01517-624517,charlesetta_erm@gmail.com,http://www.cainjohnmesq.co.uk
Corrinne,Jaret,Sound Vision Corp,2150 Morley St,Dee Ward,Dumfries and Galloway,DG8 7DE,01625-932209,01642-322954,corrinne_jaret@gmail.com,http://www.soundvisioncorp.co.uk
Niesha,Bruch,Rowley/hansell Petetin,24 Bolton St,"Broxburn, Uphall and Winchburg",West Lothian,EH52 5TL,01874-856950,01342-793603,niesha.bruch@yahoo.com,http://www.rowleyhansellpetetin.co.uk
Rueben,Gastellum,Industrial Engineering Assocs,4 Forrest St,Weston-Super-Mare,North Somerset,BS23 3HG,01976-755279,01956-535511,rueben_gastellum@gastellum.co.uk,http://www.industrialengineeringassocs.co.uk
Michell,Throssell,Weiss Spirt & Guyer,89 Noon St,Carbrooke,Norfolk,IP25 6JQ,01967-580851,01672-496478,mthrossell@throssell.co.uk,http://www.weissspirtguyer.co.uk
Edgar,Kanne,"Crowan, Kenneth W Esq",99 Guthrie St,New Milton,Hampshire,BH25 5DF,01326-532337,01666-638176,edgar.kanne@yahoo.com,http://www.crowankennethwesq.co.uk
Dewitt,Julio,Rittenhouse Motor Co,7 Richmond St,Parkham,Devon,EX39 5DJ,01253-528327,01241-964675,dewitt.julio@hotmail.com,http://www.rittenhousemotorco.co.uk
Edgar,KanneClone,"Crowan, Kenneth W Esq",99 Guthrie St,New Milton,Hampshire,BH25 5DF,01326-532337,01666-638176,edgar.kanne@yahoo.com,http://www.crowankennethwesq.co.uk
"""
df = pd.read_csv(io.StringIO(csv))
df.loc[df["email"] > "s", "web"]  # <-- should output row-to-row arrows



# https://pandas.pydata.org/docs/reference/api/pandas.DataFrame.stack.html
import pandas as pd

multicol2 = pd.MultiIndex.from_tuples([('weight', 'kg'),
                                       ('height', 'm')])
df_multi_level_cols2 = pd.DataFrame([[1.0, 2.0], [3.0, 4.0]],
                                    index=['cat', 'dog'],
                                    columns=multicol2)

df_multi_level_cols2.stack([0, 1])


"""
put an empty __init__.py inside of every directory you want to include in the
wheel

to build a wheel, run this command, (important: make sure to rm -rf the build
cache):

rm -rf build/; python setup.py bdist_wheel
"""
# type: ignore

from setuptools import setup, find_packages

version = "2.0.0.dev2"

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

# note that we don't put requirements.txt into the install_requires list
# since we bundle the dependencies directly into the wheel.

setup(
    name="pandas_tutor",
    version=version,
    packages=find_packages(),
    package_data={
        "": ["*.golden"]
    },  # add all test .golden files into package along with .py files
    include_package_data=True,
)



# tests that all nullish values are serialized to JSON null
# adapted from https://pandas.pydata.org/docs/user_guide/missing_data.html
import pandas as pd
import numpy as np

np.random.seed(42)

df = pd.DataFrame(
    np.random.randn(5, 3),
    index=["a", "c", "e", "f", "h"],
    columns=["one", "two", "three"],
)
df["four"] = "bar"
df["five"] = df["one"] > 0

df2 = df.copy()
df2["timestamp"] = pd.Timestamp("20120101")
df2.loc[["a", "c", "h"], ["one", "timestamp"]] = np.nan

# >>> df2
#         one       two     three four   five  timestamp
# a       NaN -0.282863 -1.509059  bar   True        NaT
# c       NaN  1.212112 -0.173215  bar  False        NaT
# e  0.119209 -1.044236 -0.861849  bar   True 2012-01-01
# f -2.104569 -0.494929  1.071804  bar  False 2012-01-01
# h       NaN -0.706771 -1.039575  bar   True        NaT

df2.sort_values('two')



import pandas as pd
import io

csv = """
Name,Sex,Count,Year
Noah,M,18252,2020
Julius,M,960,2020
Karen,M,6,2020
Karen,F,325,2020
Noah,F,305,2020
"""

csv2 = """
nyt_name,category
Karen,boomer
Julius,mythology
Freya,mythology
"""

baby = pd.read_csv(io.StringIO(csv))
nyt = pd.read_csv(io.StringIO(csv2))

baby2 = baby.set_index("Name")

# when right doesn't use index, we can get duplicate index vals, so we don't
# handle
baby2.merge(nyt, left_index=True, right_on="nyt_name")



"""A collection of functions to summarize object information.

This module provides several function which will help you to analyze object
information which was gathered. Often it is sufficient to work with aggregated
data instead of handling the entire set of existing objects. For example can a
memory leak identified simple based on the number and size of existing objects.

A summary contains information about objects in a table-like manner.
Technically, it is a list of lists. Each of these lists represents a row,
whereas the first column reflects the object type, the second column the number
of objects, and the third column the size of all these objects. This allows a
simple table-like output like the	following:

=============  ============  =============
       types     # objects     total size
=============  ============  =============
<type 'dict'>             2            560
 <type 'str'>             3            126
 <type 'int'>             4             96
<type 'long'>             2             66
<type 'list'>             1             40
=============  ============  =============

Another advantage of summaries is that they influence the system you analyze
only to a minimum. Working with references to existing objects will keep these
objects alive. Most of the times this is no desired behavior (as it will have
an impact on the observations). Using summaries reduces this effect greatly.

output representation
---------------------

The output representation of types is defined in summary.representations.
Every type defined in this dictionary will be represented as specified. Each
definition has a list of different representations. The later a representation
appears in this list, the higher its verbosity level. From types which are not
defined in summary.representations the default str() representation will be
used.

Per default, summaries will use the verbosity level 1 for any encountered type.
The reason is that several computations are done with summaries and rows have
to remain comparable. Therefore information which reflect an objects state,
e.g. the current line number of a frame, should not be included. You may add
more detailed information at higher verbosity levels than 1.
"""

import re
import sys
import types

from pympler.util import stringutils
from sys import getsizeof

representations = {}


def _init_representations():
    global representations
    if sys.hexversion < 0x2040000:
        classobj = [
            lambda c: "classobj(%s)" % repr(c),
        ]
        representations[types.ClassType] = classobj
        instance = [
            lambda f: "instance(%s)" % repr(f.__class__),
        ]
        representations[types.InstanceType] = instance
        instancemethod = [
            lambda i: "instancemethod (%s)" % (repr(i.im_func)),
            lambda i: "instancemethod (%s, %s)" % (repr(i.im_class),
                                                   repr(i.im_func)),
        ]
        representations[types.MethodType] = instancemethod
    frame = [
        lambda f: "frame (codename: %s)" % (f.f_code.co_name),
        lambda f: "frame (codename: %s, codeline: %s)" %
                  (f.f_code.co_name, f.f_code.co_firstlineno),
        lambda f: "frame (codename: %s, filename: %s, codeline: %s)" %
                  (f.f_code.co_name, f.f_code.co_filename,
                   f.f_code.co_firstlineno)
    ]
    representations[types.FrameType] = frame
    _dict = [
        lambda d: str(type(d)),
        lambda d: "dict, len=%s" % len(d),
    ]
    representations[dict] = _dict
    function = [
        lambda f: "function (%s)" % f.__name__,
        lambda f: "function (%s.%s)" % (f.__module, f.__name__),
    ]
    representations[types.FunctionType] = function
    _list = [
        lambda l: str(type(l)),
        lambda l: "list, len=%s" % len(l)
    ]
    representations[list] = _list
    module = [lambda m: "module(%s)" % getattr(m, '__name__', getattr(m, '__file__', 'nameless, id: %d' % id(m)))]
    representations[types.ModuleType] = module
    _set = [
        lambda s: str(type(s)),
        lambda s: "set, len=%s" % len(s)
    ]
    representations[set] = _set

_init_representations()


def summarize(objects):
    """Summarize an objects list.

    Return a list of lists, whereas each row consists of::
      [str(type), number of objects of this type, total size of these objects].

    No guarantee regarding the order is given.

    """
    count = {}
    total_size = {}
    for o in objects:
        otype = _repr(o)
        if otype in count:
            count[otype] += 1
            total_size[otype] += getsizeof(o)
        else:
            count[otype] = 1
            total_size[otype] = getsizeof(o)
    rows = []
    for otype in count:
        rows.append([otype, count[otype], total_size[otype]])
    return rows


def get_diff(left, right):
    """Get the difference of two summaries.

    Subtracts the values of the right summary from the values of the left
    summary.
    If similar rows appear on both sides, the are included in the summary with
    0 for number of elements and total size.
    If the number of elements of a row of the diff is 0, but the total size is
    not, it means that objects likely have changed, but not there number, thus
    resulting in a changed size.

    """
    res = []

    right_by_key = dict((r[0], r) for r in right)
    left_by_key  = dict((r[0], r) for r in left)

    keys = set(right_by_key)
    keys.update(left_by_key)

    for key in keys:
        r = right_by_key.get(key)
        l = left_by_key.get(key)
        if l and r:
            res.append([key, r[1] - l[1], r[2] - l[2]])
        elif r:
            res.append(r)
        elif l:
            res.append([key, -l[1], -l[2]])
        else:
            continue # shouldn't happen
    return res

def format_(rows, limit=15, sort='size', order='descending'):
    """Format the rows as a summary.

    Keyword arguments:
    limit -- the maximum number of elements to be listed
    sort  -- sort elements by 'size', 'type', or '#'
    order -- sort 'ascending' or 'descending'
    """
    localrows = []
    for row in rows:
        localrows.append(list(row))
    # input validation
    sortby = ['type', '#', 'size']
    if sort not in sortby:
        raise ValueError("invalid sort, should be one of" + str(sortby))
    orders = ['ascending', 'descending']
    if order not in orders:
        raise ValueError("invalid order, should be one of" + str(orders))
    # sort rows
    if sortby.index(sort) == 0:
        if order == "ascending":
            localrows.sort(key=lambda x: _repr(x[0]))
        elif order == "descending":
            localrows.sort(key=lambda x: _repr(x[0]), reverse=True)
    else:
        if order == "ascending":
            localrows.sort(key=lambda x: x[sortby.index(sort)])
        elif order == "descending":
            localrows.sort(key=lambda x: x[sortby.index(sort)], reverse=True)
    # limit rows
    localrows = localrows[0:limit]
    for row in localrows:
        row[2] = stringutils.pp(row[2])
    # print rows
    localrows.insert(0, ["types", "# objects", "total size"])
    return _format_table(localrows)


def _format_table(rows, header=True):
    """Format a list of lists as a pretty table.

    Keyword arguments:
    header -- if True the first row is treated as a table header

    inspired by http://aspn.activestate.com/ASPN/Cookbook/Python/Recipe/267662
    """
    border = "="
    # vertical delimiter
    vdelim = " | "
    # padding nr. of spaces are left around the longest element in the
    # column
    padding = 1
    # may be left,center,right
    justify = 'right'
    justify = {'left': str.ljust,
               'center': str.center,
               'right': str.rjust}[justify.lower()]
    # calculate column widths (longest item in each col
    # plus "padding" nr of spaces on both sides)
    cols = zip(*rows)
    colWidths = [max([len(str(item)) + 2 * padding for item in col])
                 for col in cols]
    borderline = vdelim.join([w * border for w in colWidths])
    for row in rows:
        yield vdelim.join([justify(str(item), width)
                           for (item, width) in zip(row, colWidths)])
        if header:
            yield borderline
            header = False


def print_(rows, limit=15, sort='size', order='descending'):
    """Print the rows as a summary.

    Keyword arguments:
    limit -- the maximum number of elements to be listed
    sort  -- sort elements by 'size', 'type', or '#'
    order -- sort 'ascending' or 'descending'

    """
    for line in format_(rows, limit=limit, sort=sort, order=order):
        print(line)


# regular expressions used by _repr to replace default type representations
type_repr = re.compile(r"^<(type|class) '(\S+)'>$")
address = re.compile(r' at 0x[0-9a-f]+')


def _repr(o, verbosity=1):
    """Get meaning object representation.

    This function should be used when the simple str(o) output would result in
    too general data. E.g. "<type 'instance'" is less meaningful than
    "instance: Foo".

    Keyword arguments:
    verbosity -- if True the first row is treated as a table header

    """
    res = ""

    t = type(o)
    if (verbosity == 0) or (t not in representations):
        res = str(t)
    else:
        verbosity -= 1
        if len(representations[t]) < verbosity:
            verbosity = len(representations[t]) - 1
        res = representations[t][verbosity](o)

    res = address.sub('', res)
    res = type_repr.sub(r'\2', res)

    return res


def _traverse(summary, function, *args):
    """Traverse all objects of a summary and call function with each as a
    parameter.

    Using this function, the following objects will be traversed:
    - the summary
    - each row
    - each item of a row
    """
    function(summary, *args)
    for row in summary:
        function(row, *args)
        for item in row:
            function(item, *args)


def _subtract(summary, o):
    """Remove object o from the summary by subtracting it's size."""
    found = False
    row = [_repr(o), 1, getsizeof(o)]
    for r in summary:
        if r[0] == row[0]:
            (r[1], r[2]) = (r[1] - row[1], r[2] - row[2])
            found = True
    if not found:
        summary.append([row[0], -row[1], -row[2]])
    return summary


def _sweep(summary):
    """Remove all rows in which the total size and the total number of
    objects is zero.

    """
    return [row for row in summary if ((row[2] != 0) or (row[1] != 0))]



"""
runs pandas_tutor backend

Usage:
    pandas_tutor FILE ... [--output] [--parse_only] [--parse_log]
    pandas_tutor -c CODE

Options:
    -o --output      # Outputs specs to files named {input_file}.golden
    -p --parse_only  # Outputs parsed code rather than full spec
    -l --parse_log   # Outputs parse debug output
    -c --code        # Code as a string (instead of a file)
"""

from pathlib import Path

from docopt import docopt

from .diagram import OutputSpec
from .parse import parse, parse_as_json, test_logger
from .run import run
from .serialize import serialize


def make_tutor_spec(code: str) -> str:
    """oh yeah, it's all coming together"""
    root = parse(code)
    eval_results = run(root)
    explanation = serialize(eval_results)
    spec = OutputSpec(code=code, explanation=explanation)
    return spec.to_json()


def make_tutor_spec_ipython(code: str, ipython_shell) -> str:
    """
    when we run in ipython, we need to execute code using ipython's namespace
    """
    root = parse(code)
    eval_results = run(root, ipython_shell)
    explanation = serialize(eval_results)
    spec = OutputSpec(code=code, explanation=explanation)
    return spec.to_json()


def make_tutor_spec_py(code: str) -> OutputSpec:
    """Keeps serialized output as a Python object for testing"""
    root = parse(code)
    eval_results = run(root)
    explanation = serialize(eval_results)
    spec = OutputSpec(code=code, explanation=explanation)
    return spec


def spec_from_file(filename: str, spec_fn=make_tutor_spec) -> str:
    code = Path(filename).read_text()
    return spec_fn(code)


def write_spec_to_file(spec: str, out: Path) -> None:
    print(f"Writing {out}")
    with out.open("w") as f:
        f.write(spec)


if __name__ == "__main__":
    doc = __doc__ or ""
    args = docopt(doc, version="1.0")

    spec_fn = (
        test_logger
        if args["--parse_log"]
        else parse_as_json
        if args["--parse_only"]
        else make_tutor_spec
    )

    if args["--code"]:
        code = args["CODE"]
        print(spec_fn(code))
    if not args["--output"]:
        for filename in args["FILE"]:
            print(spec_from_file(filename, spec_fn)) # type: ignore
    else:
        for filename in args["FILE"]:
            spec = spec_from_file(filename, spec_fn) # type: ignore
            out_filename = Path(filename + ".golden")
            write_spec_to_file(spec, out_filename)



# https://pandas.pydata.org/pandas-docs/stable/user_guide/reshaping.html#reshaping-by-pivoting-dataframe-objects
import pandas as pd
import io

csv = """
A,B,C,D,E
foo,one,small,1,2
foo,one,large,2,4
foo,one,large,2,5
foo,two,small,3,5
foo,two,small,3,6
bar,one,large,4,6
bar,one,small,5,8
bar,two,small,6,9
bar,two,large,7,9
"""

df = pd.read_csv(io.StringIO(csv))

# when only columns are passed, D and E get aggregated and transposed into rows
df.pivot_table(columns="C")



# https://github.com/SamLau95/pandas_tutor/issues/72
# https://pandas.pydata.org/pandas-docs/stable/user_guide/groupby.html
import pandas as pd
import numpy as np
np.random.seed(0)

arrays = [
    ["bar", "bar", "baz", "baz", "foo", "foo", "qux", "qux"],
    ["one", "two", "one", "two", "one", "two", "one", "two"],
]
index = pd.MultiIndex.from_arrays(arrays, names=["first", "second"])

df = pd.DataFrame({"A": [1, 1, 1, 1, 2, 2, 3, 3], "B": np.arange(8)}, index=index)

df.groupby([pd.Grouper(level=1), "A"]).mean()
